#    copyright (c) 2016 Industrial Technology Research Institute.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""DISCO Backup Service Implementation."""

import json

import requests
import six


class DiscoApi(object):
    """Class for all the requests to Disco API."""

    def __init__(self, ip, port):
        """Init client."""
        # Rest related variables
        self.req_headers = {'Content-type': 'application/json'}
        prefix_vars = {'server_ip': ip,
                       'server_port': port,
                       'api_prefix': 'RM-REST-Server/disco'}
        self.request_prefix = ("http://%(server_ip)s:%(server_port)s"
                               "/%(api_prefix)s") % prefix_vars
        self.prefix_var = {'req_prefix': self.request_prefix}

    def volumeCreate(self, volume_name, size):
        """Create a DISCO volume."""
        params = {'volumeName': volume_name, 'volumeSize': size,
                  'backupPolicyId': -1}
        data = json.dumps(params,
                          sort_keys=True,
                          indent=4,
                          separators=(',', ': '))
        request = ("%(req_prefix)s/volume" % self.prefix_var)
        r = requests.post(request, data, headers=self.req_headers)
        return r.json()

    def volumeDelete(self, volume_id):
        """Delete the temporary volume."""
        request_vars = {'req_prefix': self.request_prefix,
                        'volume_id': six.text_type(volume_id)}
        request = ("%(req_prefix)s/volume/%(volume_id)s") % request_vars
        r = requests.delete(request)
        return r.json()

    def volumeExtend(self, vol_id, size):
        """Extend DISCO volume."""
        params = {'volumeSize': six.text_type(size),
                  'volumeId': six.text_type(vol_id)}
        data = json.dumps(params,
                          sort_keys=True,
                          indent=4,
                          separators=(',', ': '))
        request = ("%(req_prefix)s/volume/extend" % self.prefix_var)
        r = requests.put(request, data, headers=self.req_headers)
        return r.json()

    def volumeDetail(self, volume_id):
        """Get volume information of the destination DISCO volume."""
        request_vars = {'req_prefix': self.request_prefix,
                        'vol_id': six.text_type(volume_id)}
        request = ("%(req_prefix)s/volume/%(vol_id)s") % request_vars
        r = requests.get(request)
        volume_info = r.json()
        return volume_info

    def volumeDetailByName(self, volume_name):
        """Get volume information of the DISCO volume."""
        request_vars = {'req_prefix': self.request_prefix,
                        'volume_name': six.text_type(volume_name)}
        request = ("%(req_prefix)s/volume?name=%(volume_name)s") % request_vars
        r = requests.get(request)
        return r.json()

    def volumeClone(self, volume_id, volume_name):
        """Clone a DISCO volume."""
        params = {'volumeName': volume_name, 'volumeId': volume_id}
        data = json.dumps(params,
                          sort_keys=True,
                          indent=4,
                          separators=(',', ': '))
        request = ("%(req_prefix)s/clone" % self.prefix_var)
        r = requests.post(request, data, headers=self.req_headers)
        return r.json()

    def cloneDetail(self, clone_id, clone_name):
        """Get detail of the clone."""
        request_vars = {'req_prefix': self.request_prefix,
                        'clone_name': clone_name,
                        'clone_id': six.text_type(clone_id)}
        request = ("%(req_prefix)s/clone?cloneId=%(clone_id)s&"
                   "name=%(clone_name)s") % request_vars
        r = requests.get(request)
        return r.json()

    def snapshotCreate(self, disco_volume_id, reserve_days, zone_id=None,
                       description=None):
        """Take a snapshot of the volume."""
        params = {'volumeId': disco_volume_id,
                  'reserveDays': reserve_days,
                  'description': description}
        data = json.dumps(params, sort_keys=True, indent=4,
                          separators=(',', ': '))

        request = ("%(req_prefix)s/snapshot" % self.prefix_var)
        r = requests.post(request, data, headers=self.req_headers)
        return r.json()

    def snapshotDelete(self, snapshot_id):
        """Delete a snapshot."""
        request_vars = {'req_prefix': self.request_prefix,
                        'snapshot_id': six.text_type(snapshot_id)}
        request = ("%(req_prefix)s/snapshot/%(snapshot_id)s") % request_vars
        r = requests.delete(request)
        return r.json()

    def snapshotDetail(self, snapshot_id):
        """Monitor end of the snapshot."""
        request_vars = {'req_prefix': self.request_prefix,
                        'snapshot_id': snapshot_id}
        request = ("%(req_prefix)s/snapshot/%(snapshot_id)s") % request_vars
        r = requests.get(request)
        return r.json()

    def restoreFromSnapshot(self, snapshot_id, volume_name, zone_id,
                            description, volume_id):
        """restore a snapshot of into a volume."""
        params = {'snapshotId': snapshot_id,
                  'volumeName': volume_name,
                  'zone_id': zone_id,
                  'description': "local restore snapshot",
                  'volumeId': volume_id}
        data = json.dumps(params,
                          sort_keys=True,
                          indent=4,
                          separators=(',', ': '))
        request = ("%(req_prefix)s/restore" % self.prefix_var)
        r = requests.post(request, data, headers=self.req_headers)
        return r.json()

    def restoreDetail(self, restore_id):
        """Monitor end of the restore."""
        request_vars = {'req_prefix': self.request_prefix,
                        'restore_id': restore_id}
        request = ("%(req_prefix)s/restore/%(restore_id)s") % request_vars
        r = requests.get(request)
        return r.json()

    def systemInformationList(self):
        """Get the list of the system information."""
        request = ("%(req_prefix)s/systemInformationList") % self.prefix_var
        r = requests.get(request)
        return r.json()



# Copyright (c) 2015 Alex Meade.  All rights reserved.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Mock unit tests for the NetApp E-series iscsi driver."""

import copy

import mock
import six

from cinder import exception
from cinder.objects import fields
from cinder import test
from cinder.tests.unit.volume.drivers.netapp.eseries \
    import fakes as eseries_fakes
from cinder.volume.drivers.netapp.eseries import host_mapper
from cinder.volume.drivers.netapp.eseries import utils


def get_fake_volume():
    return {
        'id': '114774fb-e15a-4fae-8ee2-c9723e3645ef', 'size': 1,
        'volume_name': 'lun1', 'host': 'hostname@backend#DDP',
        'os_type': 'linux', 'provider_location': 'lun1',
        'name_id': '114774fb-e15a-4fae-8ee2-c9723e3645ef',
        'provider_auth': 'provider a b', 'project_id': 'project',
        'display_name': None, 'display_description': 'lun1',
        'volume_type_id': None, 'migration_status': None, 'attach_status':
        fields.VolumeAttachStatus.DETACHED, "status": "available"
    }

FAKE_MAPPINGS = [{u'lun': 1}]

FAKE_USED_UP_MAPPINGS = [{u'lun': n} for n in range(256)]

FAKE_USED_UP_LUN_ID_DICT = {n: 1 for n in range(256)}

FAKE_UNUSED_LUN_ID = set([])

FAKE_USED_LUN_ID_DICT = ({0: 1, 1: 1})

FAKE_USED_LUN_IDS = [1, 2]

FAKE_SINGLE_USED_LUN_ID = 1

FAKE_USED_UP_LUN_IDS = range(256)


class NetAppEseriesHostMapperTestCase(test.TestCase):
    def setUp(self):
        super(NetAppEseriesHostMapperTestCase, self).setUp()

        self.client = eseries_fakes.FakeEseriesClient()

    def test_unmap_volume_from_host_volume_mapped_to_host(self):
        fake_eseries_volume = copy.deepcopy(eseries_fakes.VOLUME)
        fake_eseries_volume['listOfMappings'] = [
            eseries_fakes.VOLUME_MAPPING
        ]
        self.mock_object(self.client, 'list_volumes',
                         return_value=[fake_eseries_volume])
        self.mock_object(self.client, 'delete_volume_mapping')

        host_mapper.unmap_volume_from_host(self.client, get_fake_volume(),
                                           eseries_fakes.HOST,
                                           eseries_fakes.VOLUME_MAPPING)

        self.assertTrue(self.client.delete_volume_mapping.called)

    def test_unmap_volume_from_host_volume_mapped_to_different_host(self):
        fake_eseries_volume = copy.deepcopy(eseries_fakes.VOLUME)
        # Mapped to host 1
        fake_eseries_volume['listOfMappings'] = [
            eseries_fakes.VOLUME_MAPPING
        ]
        self.mock_object(self.client, 'list_volumes',
                         return_value=[fake_eseries_volume])
        self.mock_object(self.client, 'delete_volume_mapping')
        self.mock_object(self.client, 'get_host_group',
                         side_effect=exception.NotFound)

        err = self.assertRaises(exception.NetAppDriverException,
                                host_mapper.unmap_volume_from_host,
                                self.client, get_fake_volume(),
                                eseries_fakes.HOST_2,
                                eseries_fakes.VOLUME_MAPPING)
        self.assertIn("not currently mapped to host", six.text_type(err))

    def test_unmap_volume_from_host_volume_mapped_to_host_group_but_not_host(
            self):
        """Test volume mapped to host not in specified host group.

        Ensure an error is raised if the specified host is not in the
        host group the volume is mapped to.
        """
        fake_eseries_volume = copy.deepcopy(eseries_fakes.VOLUME)
        fake_volume_mapping = copy.deepcopy(eseries_fakes.VOLUME_MAPPING)
        fake_volume_mapping['mapRef'] = eseries_fakes.MULTIATTACH_HOST_GROUP[
            'clusterRef']
        fake_eseries_volume['listOfMappings'] = [fake_volume_mapping]
        self.mock_object(self.client, 'list_volumes',
                         return_value=[fake_eseries_volume])
        fake_host = copy.deepcopy(eseries_fakes.HOST)
        fake_host['clusterRef'] = utils.NULL_REF
        self.mock_object(self.client, 'list_hosts',
                         return_value=[fake_host])

        err = self.assertRaises(exception.NetAppDriverException,
                                host_mapper.unmap_volume_from_host,
                                self.client, get_fake_volume(),
                                fake_host,
                                fake_volume_mapping)
        self.assertIn("not currently mapped to host", six.text_type(err))

    def test_unmap_volume_from_host_volume_mapped_to_multiattach_host_group(
            self):
        fake_eseries_volume = copy.deepcopy(eseries_fakes.VOLUME)
        fake_volume_mapping = copy.deepcopy(eseries_fakes.VOLUME_MAPPING)
        fake_volume_mapping['mapRef'] = eseries_fakes.MULTIATTACH_HOST_GROUP[
            'clusterRef']
        fake_eseries_volume['listOfMappings'] = [fake_volume_mapping]
        self.mock_object(self.client, 'delete_volume_mapping')
        self.mock_object(self.client, 'list_volumes',
                         return_value=[fake_eseries_volume])
        fake_volume = get_fake_volume()
        fake_volume['status'] = 'detaching'

        host_mapper.unmap_volume_from_host(self.client, fake_volume,
                                           eseries_fakes.HOST,
                                           fake_volume_mapping)

        self.assertTrue(self.client.delete_volume_mapping.called)

    def test_unmap_volume_from_host_volume_mapped_to_multiattach_host_group_and_migrating(  # noqa
            self):
        fake_eseries_volume = copy.deepcopy(eseries_fakes.VOLUME)
        fake_volume_mapping = copy.deepcopy(eseries_fakes.VOLUME_MAPPING)
        fake_volume_mapping['mapRef'] = eseries_fakes.MULTIATTACH_HOST_GROUP[
            'clusterRef']
        fake_eseries_volume['listOfMappings'] = [fake_volume_mapping]
        self.mock_object(self.client, 'delete_volume_mapping')
        self.mock_object(self.client, 'list_volumes',
                         return_value=[fake_eseries_volume])
        fake_volume = get_fake_volume()
        fake_volume['status'] = 'in-use'

        host_mapper.unmap_volume_from_host(self.client, fake_volume,
                                           eseries_fakes.HOST,
                                           fake_volume_mapping)

        self.assertFalse(self.client.delete_volume_mapping.called)

    def test_unmap_volume_from_host_volume_mapped_to_outside_host_group(self):
        """Test volume mapped to host group without host.

        Ensure we raise error when we find a volume is mapped to an unknown
        host group that does not have the host.
        """
        fake_eseries_volume = copy.deepcopy(eseries_fakes.VOLUME)
        fake_volume_mapping = copy.deepcopy(eseries_fakes.VOLUME_MAPPING)
        fake_ref = "8500000060080E500023C7340036035F515B78FD"
        fake_volume_mapping['mapRef'] = fake_ref
        fake_eseries_volume['listOfMappings'] = [fake_volume_mapping]
        self.mock_object(self.client, 'list_volumes',
                         return_value=[fake_eseries_volume])
        fake_host = copy.deepcopy(eseries_fakes.HOST)
        fake_host['clusterRef'] = utils.NULL_REF
        self.mock_object(self.client, 'list_hosts',
                         return_value=[fake_host])
        self.mock_object(self.client, 'get_host_group',
                         return_value=eseries_fakes.FOREIGN_HOST_GROUP)

        err = self.assertRaises(exception.NetAppDriverException,
                                host_mapper.unmap_volume_from_host,
                                self.client, get_fake_volume(),
                                eseries_fakes.HOST,
                                fake_volume_mapping)
        self.assertIn("unsupported host group", six.text_type(err))

    def test_unmap_volume_from_host_volume_mapped_to_outside_host_group_w_host(
            self):
        """Test volume mapped to host in unknown host group.

        Ensure we raise error when we find a volume is mapped to an unknown
        host group that has the host.
        """
        fake_eseries_volume = copy.deepcopy(eseries_fakes.VOLUME)
        fake_volume_mapping = copy.deepcopy(eseries_fakes.VOLUME_MAPPING)
        fake_ref = "8500000060080E500023C7340036035F515B78FD"
        fake_volume_mapping['mapRef'] = fake_ref
        fake_eseries_volume['clusterRef'] = fake_ref
        fake_eseries_volume['listOfMappings'] = [fake_volume_mapping]
        self.mock_object(self.client, 'list_volumes',
                         return_value=[fake_eseries_volume])
        fake_host = copy.deepcopy(eseries_fakes.HOST)
        fake_host['clusterRef'] = utils.NULL_REF
        self.mock_object(self.client, 'list_hosts', return_value=[fake_host])
        self.mock_object(self.client, 'get_host_group',
                         return_value=eseries_fakes.FOREIGN_HOST_GROUP)

        err = self.assertRaises(exception.NetAppDriverException,
                                host_mapper.unmap_volume_from_host,
                                self.client, get_fake_volume(),
                                eseries_fakes.HOST,
                                fake_volume_mapping)

        self.assertIn("unsupported host group", six.text_type(err))

    def test_map_volume_to_single_host_volume_not_mapped(self):
        self.mock_object(self.client, 'create_volume_mapping',
                         return_value=eseries_fakes.VOLUME_MAPPING)

        host_mapper.map_volume_to_single_host(self.client, get_fake_volume(),
                                              eseries_fakes.VOLUME,
                                              eseries_fakes.HOST,
                                              None,
                                              False)

        self.assertTrue(self.client.create_volume_mapping.called)

    def test_map_volume_to_single_host_volume_already_mapped_to_target_host(
            self):
        """Should be a no-op"""
        self.mock_object(self.client, 'create_volume_mapping')

        host_mapper.map_volume_to_single_host(self.client,
                                              get_fake_volume(),
                                              eseries_fakes.VOLUME,
                                              eseries_fakes.HOST,
                                              eseries_fakes.VOLUME_MAPPING,
                                              False)

        self.assertFalse(self.client.create_volume_mapping.called)

    def test_map_volume_to_single_host_volume_mapped_to_multiattach_host_group(
            self):
        """Test map volume to a single host.

        Should move mapping to target host if volume is not migrating or
        attached(in-use). If volume is not in use then it should not require a
        mapping making it ok to sever the mapping to the host group.
        """
        fake_mapping_to_other_host = copy.deepcopy(
            eseries_fakes.VOLUME_MAPPING)
        fake_mapping_to_other_host['mapRef'] = \
            eseries_fakes.MULTIATTACH_HOST_GROUP['clusterRef']
        self.mock_object(self.client, 'move_volume_mapping_via_symbol',
                         return_value={'lun': 5})

        host_mapper.map_volume_to_single_host(self.client,
                                              get_fake_volume(),
                                              eseries_fakes.VOLUME,
                                              eseries_fakes.HOST,
                                              fake_mapping_to_other_host,
                                              False)

        self.assertTrue(self.client.move_volume_mapping_via_symbol.called)

    def test_map_volume_to_single_host_volume_mapped_to_multiattach_host_group_and_migrating(  # noqa
            self):
        """Should raise error saying multiattach not enabled"""
        fake_mapping_to_other_host = copy.deepcopy(
            eseries_fakes.VOLUME_MAPPING)
        fake_mapping_to_other_host['mapRef'] = \
            eseries_fakes.MULTIATTACH_HOST_GROUP['clusterRef']
        fake_volume = get_fake_volume()
        fake_volume['attach_status'] = fields.VolumeAttachStatus.ATTACHED

        err = self.assertRaises(exception.NetAppDriverException,
                                host_mapper.map_volume_to_single_host,
                                self.client, fake_volume,
                                eseries_fakes.VOLUME,
                                eseries_fakes.HOST,
                                fake_mapping_to_other_host,
                                False)

        self.assertIn('multiattach is disabled', six.text_type(err))

    def test_map_volume_to_single_host_volume_mapped_to_multiattach_host_group_and_attached(  # noqa
            self):
        """Should raise error saying multiattach not enabled"""
        fake_mapping_to_other_host = copy.deepcopy(
            eseries_fakes.VOLUME_MAPPING)
        fake_mapping_to_other_host['mapRef'] = \
            eseries_fakes.MULTIATTACH_HOST_GROUP['clusterRef']
        fake_volume = get_fake_volume()
        fake_volume['attach_status'] = fields.VolumeAttachStatus.ATTACHED

        err = self.assertRaises(exception.NetAppDriverException,
                                host_mapper.map_volume_to_single_host,
                                self.client, fake_volume,
                                eseries_fakes.VOLUME,
                                eseries_fakes.HOST,
                                fake_mapping_to_other_host,
                                False)

        self.assertIn('multiattach is disabled', six.text_type(err))

    def test_map_volume_to_single_host_volume_mapped_to_another_host(self):
        """Should raise error saying multiattach not enabled"""
        fake_mapping_to_other_host = copy.deepcopy(
            eseries_fakes.VOLUME_MAPPING)
        fake_mapping_to_other_host['mapRef'] = eseries_fakes.HOST_2[
            'hostRef']

        err = self.assertRaises(exception.NetAppDriverException,
                                host_mapper.map_volume_to_single_host,
                                self.client, get_fake_volume(),
                                eseries_fakes.VOLUME,
                                eseries_fakes.HOST,
                                fake_mapping_to_other_host,
                                False)

        self.assertIn('multiattach is disabled', six.text_type(err))

    def test_map_volume_to_multiple_hosts_volume_already_mapped_to_target_host(
            self):
        """Should be a no-op."""
        self.mock_object(self.client, 'create_volume_mapping')

        host_mapper.map_volume_to_multiple_hosts(self.client,
                                                 get_fake_volume(),
                                                 eseries_fakes.VOLUME,
                                                 eseries_fakes.HOST,
                                                 eseries_fakes.VOLUME_MAPPING)

        self.assertFalse(self.client.create_volume_mapping.called)

    def test_map_volume_to_multiple_hosts_volume_mapped_to_multiattach_host_group(  # noqa
            self):
        """Should ensure target host is in the multiattach host group."""
        fake_host = copy.deepcopy(eseries_fakes.HOST_2)
        fake_host['clusterRef'] = utils.NULL_REF

        fake_mapping_to_host_group = copy.deepcopy(
            eseries_fakes.VOLUME_MAPPING)
        fake_mapping_to_host_group['mapRef'] = \
            eseries_fakes.MULTIATTACH_HOST_GROUP['clusterRef']

        self.mock_object(self.client, 'set_host_group_for_host')
        self.mock_object(self.client, 'get_host_group',
                         return_value=eseries_fakes.MULTIATTACH_HOST_GROUP)

        host_mapper.map_volume_to_multiple_hosts(self.client,
                                                 get_fake_volume(),
                                                 eseries_fakes.VOLUME,
                                                 fake_host,
                                                 fake_mapping_to_host_group)

        self.assertEqual(
            1, self.client.set_host_group_for_host.call_count)

    def test_map_volume_to_multiple_hosts_volume_mapped_to_multiattach_host_group_with_lun_collision(  # noqa
            self):
        """Should ensure target host is in the multiattach host group."""
        fake_host = copy.deepcopy(eseries_fakes.HOST_2)
        fake_host['clusterRef'] = utils.NULL_REF
        fake_mapping_to_host_group = copy.deepcopy(
            eseries_fakes.VOLUME_MAPPING)
        fake_mapping_to_host_group['mapRef'] = \
            eseries_fakes.MULTIATTACH_HOST_GROUP['clusterRef']
        self.mock_object(self.client, 'set_host_group_for_host',
                         side_effect=exception.NetAppDriverException)

        self.assertRaises(exception.NetAppDriverException,
                          host_mapper.map_volume_to_multiple_hosts,
                          self.client,
                          get_fake_volume(),
                          eseries_fakes.VOLUME,
                          fake_host,
                          fake_mapping_to_host_group)

    def test_map_volume_to_multiple_hosts_volume_mapped_to_another_host(self):
        """Test that mapping moves to another host group.

        Should ensure both existing host and destination host are in
        multiattach host group and move the mapping to the host group.
        """

        existing_host = copy.deepcopy(eseries_fakes.HOST)
        existing_host['clusterRef'] = utils.NULL_REF
        target_host = copy.deepcopy(eseries_fakes.HOST_2)
        target_host['clusterRef'] = utils.NULL_REF
        self.mock_object(self.client, 'get_host', return_value=existing_host)
        self.mock_object(self.client, 'set_host_group_for_host')
        self.mock_object(self.client, 'get_host_group',
                         side_effect=exception.NotFound)
        mock_move_mapping = mock.Mock(
            return_value=eseries_fakes.VOLUME_MAPPING_TO_MULTIATTACH_GROUP)
        self.mock_object(self.client,
                         'move_volume_mapping_via_symbol',
                         mock_move_mapping)

        host_mapper.map_volume_to_multiple_hosts(self.client,
                                                 get_fake_volume(),
                                                 eseries_fakes.VOLUME,
                                                 target_host,
                                                 eseries_fakes.VOLUME_MAPPING)

        self.assertEqual(
            2, self.client.set_host_group_for_host.call_count)

        self.assertTrue(self.client.move_volume_mapping_via_symbol
                        .called)

    def test_map_volume_to_multiple_hosts_volume_mapped_to_another_host_with_lun_collision_with_source_host(  # noqa
            self):
        """Test moving source host to multiattach host group.

        Should fail attempting to move source host to multiattach host
        group and raise an error.
        """

        existing_host = copy.deepcopy(eseries_fakes.HOST)
        existing_host['clusterRef'] = utils.NULL_REF
        target_host = copy.deepcopy(eseries_fakes.HOST_2)
        target_host['clusterRef'] = utils.NULL_REF
        self.mock_object(self.client, 'get_host', return_value=existing_host)
        self.mock_object(self.client, 'set_host_group_for_host',
                         side_effect=[None, exception.NetAppDriverException])
        self.mock_object(self.client, 'get_host_group',
                         side_effect=exception.NotFound)
        mock_move_mapping = mock.Mock(
            return_value=eseries_fakes.VOLUME_MAPPING_TO_MULTIATTACH_GROUP)
        self.mock_object(self.client,
                         'move_volume_mapping_via_symbol',
                         mock_move_mapping)

        self.assertRaises(exception.NetAppDriverException,
                          host_mapper.map_volume_to_multiple_hosts,
                          self.client,
                          get_fake_volume(),
                          eseries_fakes.VOLUME,
                          target_host,
                          eseries_fakes.VOLUME_MAPPING)

    def test_map_volume_to_multiple_hosts_volume_mapped_to_another_host_with_lun_collision_with_dest_host(  # noqa
            self):
        """Test moving destination host to multiattach host group.

        Should fail attempting to move destination host to multiattach host
        group and raise an error.
        """

        existing_host = copy.deepcopy(eseries_fakes.HOST)
        existing_host['clusterRef'] = utils.NULL_REF
        target_host = copy.deepcopy(eseries_fakes.HOST_2)
        target_host['clusterRef'] = utils.NULL_REF
        self.mock_object(self.client, 'get_host', return_value=existing_host)
        self.mock_object(self.client, 'set_host_group_for_host',
                         side_effect=[exception.NetAppDriverException, None])
        self.mock_object(self.client, 'get_host_group',
                         side_effect=exception.NotFound)
        mock_move_mapping = mock.Mock(
            return_value=eseries_fakes.VOLUME_MAPPING_TO_MULTIATTACH_GROUP)
        self.mock_object(self.client,
                         'move_volume_mapping_via_symbol',
                         mock_move_mapping)

        self.assertRaises(exception.NetAppDriverException,
                          host_mapper.map_volume_to_multiple_hosts,
                          self.client,
                          get_fake_volume(),
                          eseries_fakes.VOLUME,
                          target_host,
                          eseries_fakes.VOLUME_MAPPING)

    def test_map_volume_to_multiple_hosts_volume_mapped_to_foreign_host_group(
            self):
        """Test a target when the host is in a foreign host group.

        Should raise an error stating the volume is mapped to an
        unsupported host group.
        """
        fake_eseries_volume = copy.deepcopy(eseries_fakes.VOLUME)
        fake_volume_mapping = copy.deepcopy(eseries_fakes.VOLUME_MAPPING)
        fake_ref = "8500000060080E500023C7340036035F515B78FD"
        fake_volume_mapping['mapRef'] = fake_ref
        self.mock_object(self.client, 'list_volumes',
                         return_value=[fake_eseries_volume])
        fake_host = copy.deepcopy(eseries_fakes.HOST)
        fake_host['clusterRef'] = utils.NULL_REF
        self.mock_object(self.client, 'get_host_group',
                         return_value=eseries_fakes.FOREIGN_HOST_GROUP)

        err = self.assertRaises(exception.NetAppDriverException,
                                host_mapper.map_volume_to_multiple_hosts,
                                self.client,
                                get_fake_volume(),
                                eseries_fakes.VOLUME,
                                fake_host,
                                fake_volume_mapping)
        self.assertIn("unsupported host group", six.text_type(err))

    def test_map_volume_to_multiple_hosts_volume_mapped_to_host_in_foreign_host_group(  # noqa
            self):
        """Test a target when the host is in a foreign host group.

        Should raise an error stating the volume is mapped to a
        host that is in an unsupported host group.
        """
        fake_eseries_volume = copy.deepcopy(eseries_fakes.VOLUME)
        fake_volume_mapping = copy.deepcopy(eseries_fakes.VOLUME_MAPPING)
        fake_host = copy.deepcopy(eseries_fakes.HOST_2)
        fake_host['clusterRef'] = eseries_fakes.FOREIGN_HOST_GROUP[
            'clusterRef']
        fake_volume_mapping['mapRef'] = fake_host['hostRef']
        fake_eseries_volume['listOfMappings'] = [fake_volume_mapping]
        self.mock_object(self.client, 'list_volumes',
                         return_value=[fake_eseries_volume])
        self.mock_object(self.client, 'get_host', return_value=fake_host)
        self.mock_object(self.client, 'get_host_group',
                         side_effect=[eseries_fakes.FOREIGN_HOST_GROUP])

        err = self.assertRaises(exception.NetAppDriverException,
                                host_mapper.map_volume_to_multiple_hosts,
                                self.client,
                                get_fake_volume(),
                                eseries_fakes.VOLUME,
                                eseries_fakes.HOST,
                                fake_volume_mapping)

        self.assertIn("unsupported host group", six.text_type(err))

    def test_map_volume_to_multiple_hosts_volume_target_host_in_foreign_host_group(  # noqa
            self):
        """Test a target when the host is in a foreign host group.

        Should raise an error stating the target host is in an
        unsupported host group.
        """
        fake_eseries_volume = copy.deepcopy(eseries_fakes.VOLUME)
        fake_volume_mapping = copy.deepcopy(eseries_fakes.VOLUME_MAPPING)
        fake_host = copy.deepcopy(eseries_fakes.HOST_2)
        fake_host['clusterRef'] = eseries_fakes.FOREIGN_HOST_GROUP[
            'clusterRef']
        self.mock_object(self.client, 'list_volumes',
                         return_value=[fake_eseries_volume])
        self.mock_object(self.client, 'get_host',
                         return_value=eseries_fakes.HOST)
        self.mock_object(self.client, 'get_host_group',
                         side_effect=[eseries_fakes.FOREIGN_HOST_GROUP])

        err = self.assertRaises(exception.NetAppDriverException,
                                host_mapper.map_volume_to_multiple_hosts,
                                self.client,
                                get_fake_volume(),
                                eseries_fakes.VOLUME,
                                fake_host,
                                fake_volume_mapping)

        self.assertIn("unsupported host group", six.text_type(err))

    def test_get_unused_lun_ids(self):
        unused_lun_ids = host_mapper._get_unused_lun_ids(FAKE_MAPPINGS)
        self.assertEqual(set(range(2, 256)), unused_lun_ids)

    def test_get_unused_lun_id_counter(self):
        used_lun_id_count = host_mapper._get_used_lun_id_counter(
            FAKE_MAPPINGS)
        self.assertEqual(FAKE_USED_LUN_ID_DICT, used_lun_id_count)

    def test_get_unused_lun_ids_used_up_luns(self):
        unused_lun_ids = host_mapper._get_unused_lun_ids(
            FAKE_USED_UP_MAPPINGS)
        self.assertEqual(FAKE_UNUSED_LUN_ID, unused_lun_ids)

    def test_get_lun_id_counter_used_up_luns(self):
        used_lun_ids = host_mapper._get_used_lun_id_counter(
            FAKE_USED_UP_MAPPINGS)
        self.assertEqual(FAKE_USED_UP_LUN_ID_DICT, used_lun_ids)

    def test_host_not_full(self):
        fake_host = copy.deepcopy(eseries_fakes.HOST)
        self.assertFalse(host_mapper._is_host_full(self.client, fake_host))

    def test_host_full(self):
        fake_host = copy.deepcopy(eseries_fakes.HOST)
        self.mock_object(self.client, 'get_volume_mappings_for_host',
                         return_value=FAKE_USED_UP_MAPPINGS)
        self.assertTrue(host_mapper._is_host_full(self.client, fake_host))

    def test_get_free_lun(self):
        fake_host = copy.deepcopy(eseries_fakes.HOST)
        with mock.patch('random.sample') as mock_random:
            mock_random.return_value = [3]
            lun = host_mapper._get_free_lun(self.client, fake_host, False,
                                            [])
        self.assertEqual(3, lun)

    def test_get_free_lun_host_full(self):
        fake_host = copy.deepcopy(eseries_fakes.HOST)
        self.mock_object(host_mapper, '_is_host_full', return_value=True)
        self.assertRaises(
            exception.NetAppDriverException,
            host_mapper._get_free_lun,
            self.client, fake_host, False, FAKE_USED_UP_MAPPINGS)

    def test_get_free_lun_no_unused_luns(self):
        fake_host = copy.deepcopy(eseries_fakes.HOST)
        lun = host_mapper._get_free_lun(self.client, fake_host, False,
                                        FAKE_USED_UP_MAPPINGS)
        self.assertEqual(255, lun)

    def test_get_free_lun_no_unused_luns_host_not_full(self):
        fake_host = copy.deepcopy(eseries_fakes.HOST)
        self.mock_object(host_mapper, '_is_host_full', return_value=False)
        lun = host_mapper._get_free_lun(self.client, fake_host, False,
                                        FAKE_USED_UP_MAPPINGS)
        self.assertEqual(255, lun)

    def test_get_free_lun_no_lun_available(self):
        fake_host = copy.deepcopy(eseries_fakes.HOST_3)
        self.mock_object(self.client, 'get_volume_mappings_for_host',
                         return_value=FAKE_USED_UP_MAPPINGS)

        self.assertRaises(exception.NetAppDriverException,
                          host_mapper._get_free_lun,
                          self.client, fake_host, False,
                          FAKE_USED_UP_MAPPINGS)

    def test_get_free_lun_multiattach_enabled_no_unused_ids(self):
        fake_host = copy.deepcopy(eseries_fakes.HOST_3)
        self.mock_object(self.client, 'get_volume_mappings',
                         return_value=FAKE_USED_UP_MAPPINGS)

        self.assertRaises(exception.NetAppDriverException,
                          host_mapper._get_free_lun,
                          self.client, fake_host, True,
                          FAKE_USED_UP_MAPPINGS)

    def test_get_lun_by_mapping(self):
        used_luns = host_mapper._get_used_lun_ids_for_mappings(FAKE_MAPPINGS)
        self.assertEqual(set([0, 1]), used_luns)

    def test_get_lun_by_mapping_no_mapping(self):
        used_luns = host_mapper._get_used_lun_ids_for_mappings([])
        self.assertEqual(set([0]), used_luns)

    def test_lun_id_available_on_host(self):
        fake_host = copy.deepcopy(eseries_fakes.HOST)
        self.assertTrue(host_mapper._is_lun_id_available_on_host(
            self.client, fake_host, FAKE_UNUSED_LUN_ID))

    def test_no_lun_id_available_on_host(self):
        fake_host = copy.deepcopy(eseries_fakes.HOST_3)
        self.mock_object(self.client, 'get_volume_mappings_for_host',
                         return_value=FAKE_USED_UP_MAPPINGS)

        self.assertFalse(host_mapper._is_lun_id_available_on_host(
            self.client, fake_host, FAKE_SINGLE_USED_LUN_ID))



# Copyright 2016 Nexenta Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from oslo_log import log as logging
from oslo_utils import units

from cinder import context
from cinder import db
from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder.volume import driver
from cinder.volume.drivers.nexenta.ns5 import jsonrpc
from cinder.volume.drivers.nexenta import options
from cinder.volume.drivers.nexenta import utils

VERSION = '1.1.0'
LOG = logging.getLogger(__name__)
TARGET_GROUP_PREFIX = 'cinder-tg-'


@interface.volumedriver
class NexentaISCSIDriver(driver.ISCSIDriver):
    """Executes volume driver commands on Nexenta Appliance.

    .. code-block:: default

      Version history:
          1.0.0 - Initial driver version.
          1.1.0 - Added HTTPS support.
                  Added use of sessions for REST calls.

    """

    VERSION = VERSION

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = "Nexenta_CI"

    def __init__(self, *args, **kwargs):
        super(NexentaISCSIDriver, self).__init__(*args, **kwargs)
        self.nef = None
        # mapping of targets and groups. Groups are the keys
        self.targets = {}
        # list of volumes in target group. Groups are the keys
        self.volumes = {}
        if self.configuration:
            self.configuration.append_config_values(
                options.NEXENTA_CONNECTION_OPTS)
            self.configuration.append_config_values(
                options.NEXENTA_ISCSI_OPTS)
            self.configuration.append_config_values(
                options.NEXENTA_DATASET_OPTS)
            self.configuration.append_config_values(
                options.NEXENTA_RRMGR_OPTS)
        self.use_https = self.configuration.nexenta_use_https
        self.nef_host = self.configuration.nexenta_host
        self.nef_port = self.configuration.nexenta_rest_port
        self.nef_user = self.configuration.nexenta_user
        self.nef_password = self.configuration.nexenta_password
        self.storage_pool = self.configuration.nexenta_volume
        self.volume_group = self.configuration.nexenta_volume_group
        self.dataset_compression = (
            self.configuration.nexenta_dataset_compression)
        self.dataset_deduplication = self.configuration.nexenta_dataset_dedup
        self.dataset_description = (
            self.configuration.nexenta_dataset_description)
        self.iscsi_target_portal_port = (
            self.configuration.nexenta_iscsi_target_portal_port)

    @property
    def backend_name(self):
        backend_name = None
        if self.configuration:
            backend_name = self.configuration.safe_get('volume_backend_name')
        if not backend_name:
            backend_name = self.__class__.__name__
        return backend_name

    def do_setup(self, context):
        self.nef = jsonrpc.NexentaJSONProxy(
            self.nef_host, self.nef_port, self.nef_user,
            self.nef_password, self.use_https)
        url = 'storage/pools/%s/volumeGroups' % self.storage_pool
        data = {
            'name': self.volume_group,
            'volumeBlockSize': (
                self.configuration.nexenta_ns5_blocksize * units.Ki)
        }
        try:
            self.nef.post(url, data)
        except exception.NexentaException as e:
            if 'EEXIST' in e.args[0]:
                LOG.debug('volumeGroup already exists, skipping')
            else:
                raise

        self._fetch_volumes()

    def _fetch_volumes(self):
        url = 'san/iscsi/targets?fields=alias,name&limit=50000'
        for target in self.nef.get(url)['data']:
            tg_name = target['alias']
            if tg_name.startswith(TARGET_GROUP_PREFIX):
                self.targets[tg_name] = target['name']
                self._fill_volumes(tg_name)

    def check_for_setup_error(self):
        """Verify that the zfs volumes exist.

        :raise: :py:exc:`LookupError`
        """
        url = 'storage/pools/%(pool)s/volumeGroups/%(group)s' % {
            'pool': self.storage_pool,
            'group': self.volume_group,
        }
        try:
            self.nef.get(url)
        except exception.NexentaException:
            raise LookupError(_(
                "Dataset group %s not found at Nexenta SA"), '/'.join(
                [self.storage_pool, self.volume_group]))
        services = self.nef.get('services')
        for service in services['data']:
            if service['name'] == 'iscsit':
                if service['state'] != 'online':
                    raise exception.NexentaException(
                        'iSCSI service is not running on NS appliance')
                break

    def _get_volume_path(self, volume):
        """Return zfs volume name that corresponds given volume name."""
        return '%s/%s/%s' % (self.storage_pool, self.volume_group,
                             volume['name'])

    @staticmethod
    def _get_clone_snapshot_name(volume):
        """Return name for snapshot that will be used to clone the volume."""
        return 'cinder-clone-snapshot-%(id)s' % volume

    def create_volume(self, volume):
        """Create a zfs volume on appliance.

        :param volume: volume reference
        :return: model update dict for volume reference
        """
        url = 'storage/pools/%(pool)s/volumeGroups/%(group)s/volumes' % {
            'pool': self.storage_pool,
            'group': self.volume_group,
        }
        data = {
            'name': volume['name'],
            'volumeSize': volume['size'] * units.Gi,
            'volumeBlockSize': (
                self.configuration.nexenta_ns5_blocksize * units.Ki),
            'sparseVolume': self.configuration.nexenta_sparse
        }
        self.nef.post(url, data)

    def delete_volume(self, volume):
        """Destroy a zfs volume on appliance.

        :param volume: volume reference
        """

        url = ('storage/pools/%(pool)s/volumeGroups/%(group)s'
               '/volumes/%(name)s') % {
            'pool': self.storage_pool,
            'group': self.volume_group,
            'name': volume['name']
        }
        try:
            self.nef.delete(url)
        except exception.NexentaException as exc:
            # We assume that volume is gone
            LOG.warning('Got error trying to delete volume %(volume)s,'
                        ' assuming it is already gone: %(exc)s',
                        {'volume': volume, 'exc': exc})

    def extend_volume(self, volume, new_size):
        """Extend an existing volume.

        :param volume: volume reference
        :param new_size: volume new size in GB
        """
        LOG.info('Extending volume: %(id)s New size: %(size)s GB',
                 {'id': volume['id'], 'size': new_size})
        pool, group, name = self._get_volume_path(volume).split('/')
        url = ('storage/pools/%(pool)s/volumeGroups/%(group)s/'
               'volumes/%(name)s') % {
            'pool': pool,
            'group': group,
            'name': name
        }
        self.nef.put(url, {'volumeSize': new_size * units.Gi})

    def create_snapshot(self, snapshot):
        """Creates a snapshot.

        :param snapshot: snapshot reference
        """
        snapshot_vol = self._get_snapshot_volume(snapshot)
        LOG.info('Creating snapshot %(snap)s of volume %(vol)s', {
            'snap': snapshot['name'],
            'vol': snapshot_vol['name']
        })
        volume_path = self._get_volume_path(snapshot_vol)
        pool, group, volume = volume_path.split('/')
        url = ('storage/pools/%(pool)s/volumeGroups/%(group)s/'
               'volumes/%(volume)s/snapshots') % {
            'pool': pool,
            'group': group,
            'volume': snapshot_vol['name']
        }
        self.nef.post(url, {'name': snapshot['name']})

    def delete_snapshot(self, snapshot):
        """Delete volume's snapshot on appliance.

        :param snapshot: snapshot reference
        """
        LOG.info('Deleting snapshot: %s', snapshot['name'])
        snapshot_vol = self._get_snapshot_volume(snapshot)
        volume_path = self._get_volume_path(snapshot_vol)
        pool, group, volume = volume_path.split('/')
        url = ('storage/pools/%(pool)s/volumeGroups/%(group)s/'
               'volumes/%(volume)s/snapshots/%(snapshot)s') % {
            'pool': pool,
            'group': group,
            'volume': volume,
            'snapshot': snapshot['name']
        }
        try:
            self.nef.delete(url)
        except exception.NexentaException as exc:
            if 'EBUSY' in exc.args[0]:
                LOG.warning(
                    'Could not delete snapshot %s - it has dependencies',
                    snapshot['name'])
            else:
                LOG.warning(exc)

    def create_volume_from_snapshot(self, volume, snapshot):
        """Create new volume from other's snapshot on appliance.

        :param volume: reference of volume to be created
        :param snapshot: reference of source snapshot
        """
        LOG.info('Creating volume from snapshot: %s', snapshot['name'])
        snapshot_vol = self._get_snapshot_volume(snapshot)
        volume_path = self._get_volume_path(snapshot_vol)
        pool, group, snapshot_vol = volume_path.split('/')
        url = ('storage/pools/%(pool)s/volumeGroups/%(group)s/'
               'volumes/%(volume)s/snapshots/%(snapshot)s/clone') % {
            'pool': pool,
            'group': group,
            'volume': snapshot_vol,
            'snapshot': snapshot['name']
        }
        self.nef.post(url, {'targetPath': self._get_volume_path(volume)})
        if (('size' in volume) and (
                volume['size'] > snapshot['volume_size'])):
            self.extend_volume(volume, volume['size'])

    def create_cloned_volume(self, volume, src_vref):
        """Creates a clone of the specified volume.

        :param volume: new volume reference
        :param src_vref: source volume reference
        """
        snapshot = {'volume_name': src_vref['name'],
                    'volume_id': src_vref['id'],
                    'volume_size': src_vref['size'],
                    'name': self._get_clone_snapshot_name(volume)}
        LOG.debug('Creating temp snapshot of the original volume: '
                  '%s@%s', snapshot['volume_name'], snapshot['name'])
        self.create_snapshot(snapshot)
        try:
            self.create_volume_from_snapshot(volume, snapshot)
        except exception.NexentaException:
            LOG.error('Volume creation failed, deleting created snapshot %s',
                      '@'.join([snapshot['volume_name'], snapshot['name']]))
            try:
                self.delete_snapshot(snapshot)
            except (exception.NexentaException, exception.SnapshotIsBusy):
                LOG.warning('Failed to delete zfs snapshot %s',
                            '@'.join([snapshot['volume_name'],
                                      snapshot['name']]))
            raise

    def _get_snapshot_volume(self, snapshot):
        ctxt = context.get_admin_context()
        return db.volume_get(ctxt, snapshot['volume_id'])

    def _do_export(self, _ctx, volume):
        """Do all steps to get zfs volume exported at separate target.

        :param volume: reference of volume to be exported
        """
        volume_path = self._get_volume_path(volume)

        # Find out whether the volume is exported
        vol_map_url = 'san/lunMappings?volume=%s&fields=lun' % (
            volume_path.replace('/', '%2F'))
        data = self.nef.get(vol_map_url).get('data')
        if data:
            model_update = {}
        else:
            # Choose the best target group among existing ones
            tg_name = None
            for tg in self.volumes.keys():
                if len(self.volumes[tg]) < 20:
                    tg_name = tg
                    break
            if tg_name:
                target_name = self.targets[tg_name]
            else:
                tg_name = TARGET_GROUP_PREFIX + uuid.uuid4().hex

                # Create new target
                url = 'san/iscsi/targets'
                data = {
                    "portals": [
                        {"address": self.nef_host}
                    ],
                    'alias': tg_name
                }
                self.nef.post(url, data)

                # Get the name of just created target
                data = self.nef.get(
                    '%(url)s?fields=name&alias=%(tg_name)s' % {
                        'url': url,
                        'tg_name': tg_name
                    })['data']
                target_name = data[0]['name']

                self._create_target_group(tg_name, target_name)

                self.targets[tg_name] = target_name
                self.volumes[tg_name] = set()

            # Export the volume
            url = 'san/lunMappings'
            data = {
                "hostGroup": "all",
                "targetGroup": tg_name,
                'volume': volume_path
            }
            try:
                self.nef.post(url, data)
                self.volumes[tg_name].add(volume_path)
            except exception.NexentaException as e:
                if 'No such target group' in e.args[0]:
                    self._create_target_group(tg_name, target_name)
                    self._fill_volumes(tg_name)
                    self.nef.post(url, data)
                else:
                    raise

            # Get LUN of just created volume
            data = self.nef.get(vol_map_url).get('data')
            lun = data[0]['lun']

            provider_location = '%(host)s:%(port)s,1 %(name)s %(lun)s' % {
                'host': self.nef_host,
                'port': self.configuration.nexenta_iscsi_target_portal_port,
                'name': target_name,
                'lun': lun,
            }
            model_update = {'provider_location': provider_location}
        return model_update

    def create_export(self, _ctx, volume, connector):
        """Create new export for zfs volume.

        :param volume: reference of volume to be exported
        :return: iscsiadm-formatted provider location string
        """
        model_update = self._do_export(_ctx, volume)
        return model_update

    def ensure_export(self, _ctx, volume):
        """Recreate parts of export if necessary.

        :param volume: reference of volume to be exported
        """
        self._do_export(_ctx, volume)

    def remove_export(self, _ctx, volume):
        """Destroy all resources created to export zfs volume.

        :param volume: reference of volume to be unexported
        """
        volume_path = self._get_volume_path(volume)

        # Get ID of a LUN mapping if the volume is exported
        url = 'san/lunMappings?volume=%s&fields=id' % (
            volume_path.replace('/', '%2F'))
        data = self.nef.get(url)['data']
        if data:
            url = 'san/lunMappings/%s' % data[0]['id']
            self.nef.delete(url)
        else:
            LOG.debug('LU already deleted from appliance')

        for tg in self.volumes:
            if volume_path in self.volumes[tg]:
                self.volumes[tg].remove(volume_path)
                break

    def get_volume_stats(self, refresh=False):
        """Get volume stats.

        If 'refresh' is True, run update the stats first.
        """
        if refresh:
            self._update_volume_stats()

        return self._stats

    def _update_volume_stats(self):
        """Retrieve stats info for NexentaStor appliance."""
        LOG.debug('Updating volume stats')

        url = ('storage/pools/%(pool)s/volumeGroups/%(group)s'
               '?fields=bytesAvailable,bytesUsed') % {
            'pool': self.storage_pool,
            'group': self.volume_group,
        }
        stats = self.nef.get(url)
        total_amount = utils.str2gib_size(stats['bytesAvailable'])
        free_amount = utils.str2gib_size(
            stats['bytesAvailable'] - stats['bytesUsed'])

        location_info = '%(driver)s:%(host)s:%(pool)s/%(group)s' % {
            'driver': self.__class__.__name__,
            'host': self.nef_host,
            'pool': self.storage_pool,
            'group': self.volume_group,
        }
        self._stats = {
            'vendor_name': 'Nexenta',
            'dedup': self.dataset_deduplication,
            'compression': self.dataset_compression,
            'description': self.dataset_description,
            'driver_version': self.VERSION,
            'storage_protocol': 'iSCSI',
            'total_capacity_gb': total_amount,
            'free_capacity_gb': free_amount,
            'reserved_percentage': self.configuration.reserved_percentage,
            'QoS_support': False,
            'volume_backend_name': self.backend_name,
            'location_info': location_info,
            'iscsi_target_portal_port': self.iscsi_target_portal_port,
            'nef_url': self.nef.url
        }

    def _fill_volumes(self, tg_name):
        url = ('san/lunMappings?targetGroup=%s&fields=volume'
               '&limit=50000' % tg_name)
        self.volumes[tg_name] = {
            mapping['volume'] for mapping in self.nef.get(url)['data']}

    def _create_target_group(self, tg_name, target_name):
        # Create new target group
        url = 'san/targetgroups'
        data = {
            'name': tg_name,
            'members': [target_name]
        }
        self.nef.post(url, data)



# Copyright (c) 2016 EMC Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Contains some commonly used utility methods."""
try:
    import cookielib as cookie_lib
except ImportError:
    import http.cookiejar as cookie_lib
import json
import re
import socket

import oslo_serialization
from oslo_utils import timeutils
from oslo_utils import units
import requests
from requests import exceptions
import six
from six.moves import http_client

from cinder import exception
from cinder.i18n import _
from cinder.volume.drivers.coprhd.helpers import urihelper


PROD_NAME = 'storageos'

TIMEOUT_SEC = 20  # 20 SECONDS

global AUTH_TOKEN
AUTH_TOKEN = None

TASK_TIMEOUT = 300

URI_TASKS_BY_OPID = '/vdc/tasks/{0}'


def _decode_list(data):
    rv = []
    for item in data:
        if isinstance(item, six.text_type):
            item = item.encode('utf-8')
        elif isinstance(item, list):
            item = _decode_list(item)
        elif isinstance(item, dict):
            item = _decode_dict(item)
        rv.append(item)
    return rv


def _decode_dict(data):
    rv = {}
    for key, value in data.items():
        if isinstance(key, six.text_type):
            key = key.encode('utf-8')
        if isinstance(value, six.text_type):
            value = value.encode('utf-8')
        elif isinstance(value, list):
            value = _decode_list(value)
        elif isinstance(value, dict):
            value = _decode_dict(value)
        rv[key] = value
    return rv


def json_decode(rsp):
    """Used to decode the JSON encoded response."""

    try:
        o = json.loads(rsp, object_hook=_decode_dict)
    except ValueError:
        raise CoprHdError(CoprHdError.VALUE_ERR,
                          (_("Failed to recognize JSON payload:\n[%s]") % rsp))
    return o


def service_json_request(ip_addr, port, http_method, uri, body,
                         contenttype='application/json', customheaders=None):
    """Used to make an HTTP request and get the response.

    The message body is encoded in JSON format

    :param ip_addr: IP address or host name of the server
    :param port: port number of the server on which it
            is listening to HTTP requests
    :param http_method: one of GET, POST, PUT, DELETE
    :param uri: the request URI
    :param body: the request payload
    :returns: a tuple of two elements: (response body, response headers)
    :raises CoprHdError: in case of HTTP errors with err_code 3
    """

    SEC_AUTHTOKEN_HEADER = 'X-SDS-AUTH-TOKEN'

    headers = {'Content-Type': contenttype,
               'ACCEPT': 'application/json, application/octet-stream',
               'X-EMC-REST-CLIENT': 'TRUE'}

    if customheaders:
        headers.update(customheaders)

    try:
        protocol = "https://"
        if port == 8080:
            protocol = "http://"
        url = protocol + ip_addr + ":" + six.text_type(port) + uri

        cookiejar = cookie_lib.LWPCookieJar()
        headers[SEC_AUTHTOKEN_HEADER] = AUTH_TOKEN

        if http_method == 'GET':
            response = requests.get(url, headers=headers, verify=False,
                                    cookies=cookiejar)
        elif http_method == 'POST':
            response = requests.post(url, data=body, headers=headers,
                                     verify=False, cookies=cookiejar)
        elif http_method == 'PUT':
            response = requests.put(url, data=body, headers=headers,
                                    verify=False, cookies=cookiejar)
        elif http_method == 'DELETE':

            response = requests.delete(url, headers=headers, verify=False,
                                       cookies=cookiejar)
        else:
            raise CoprHdError(CoprHdError.HTTP_ERR,
                              (_("Unknown/Unsupported HTTP method: %s") %
                               http_method))

        if (response.status_code == http_client.OK or
                response.status_code == http_client.ACCEPTED):
            return (response.text, response.headers)

        error_msg = None
        if response.status_code == http_client.INTERNAL_SERVER_ERROR:
            response_text = json_decode(response.text)
            error_details = ""
            if 'details' in response_text:
                error_details = response_text['details']
            error_msg = (_("CoprHD internal server error. Error details: %s"),
                         error_details)
        elif response.status_code == http_client.UNAUTHORIZED:
            error_msg = _("Access forbidden: Authentication required")
        elif response.status_code == http_client.FORBIDDEN:
            error_msg = ""
            error_details = ""
            error_description = ""

            response_text = json_decode(response.text)

            if 'details' in response_text:
                error_details = response_text['details']
                error_msg = (_("%(error_msg)s Error details:"
                               " %(error_details)s"),
                             {'error_msg': error_msg,
                              'error_details': error_details
                              })
            elif 'description' in response_text:
                error_description = response_text['description']
                error_msg = (_("%(error_msg)s Error description:"
                               " %(error_description)s"),
                             {'error_msg': error_msg,
                              'error_description': error_description
                              })
            else:
                error_msg = _("Access forbidden: You don't have"
                              " sufficient privileges to perform this"
                              " operation")

        elif response.status_code == http_client.NOT_FOUND:
            error_msg = "Requested resource not found"
        elif response.status_code == http_client.METHOD_NOT_ALLOWED:
            error_msg = six.text_type(response.text)
        elif response.status_code == http_client.SERVICE_UNAVAILABLE:
            error_msg = ""
            error_details = ""
            error_description = ""

            response_text = json_decode(response.text)

            if 'code' in response_text:
                errorCode = response_text['code']
                error_msg = "Error " + six.text_type(errorCode)

            if 'details' in response_text:
                error_details = response_text['details']
                error_msg = error_msg + ": " + error_details
            elif 'description' in response_text:
                error_description = response_text['description']
                error_msg = error_msg + ": " + error_description
            else:
                error_msg = _("Service temporarily unavailable:"
                              " The server is temporarily unable to"
                              " service your request")
        else:
            error_msg = response.text
            if isinstance(error_msg, six.text_type):
                error_msg = error_msg.encode('utf-8')
        raise CoprHdError(CoprHdError.HTTP_ERR,
                          (_("HTTP code: %(status_code)s"
                             ", %(reason)s"
                             " [%(error_msg)s]") % {
                              'status_code': six.text_type(
                                  response.status_code),
                              'reason': six.text_type(
                                  response.reason),
                              'error_msg': six.text_type(
                                  error_msg)
                          }))
    except (CoprHdError, socket.error, exceptions.SSLError,
            exceptions.ConnectionError, exceptions.TooManyRedirects,
            exceptions.Timeout) as e:
        raise CoprHdError(CoprHdError.HTTP_ERR, six.text_type(e))
    # TODO(Ravi) : Either following exception should have proper message or
    # IOError should just be combined with the above statement
    except IOError as e:
        raise CoprHdError(CoprHdError.HTTP_ERR, six.text_type(e))


def is_uri(name):
    """Checks whether the name is a URI or not.

    :param name: Name of the resource
    :returns: True if name is URI, False otherwise
    """
    try:
        (urn, prod, trailer) = name.split(':', 2)
        return (urn == 'urn' and prod == PROD_NAME)
    except Exception:
        return False


def format_json_object(obj):
    """Formats JSON object to make it readable by proper indentation.

    :param obj: JSON object
    :returns: a string of  formatted JSON object
    """
    return oslo_serialization.jsonutils.dumps(obj, sort_keys=True, indent=3)


def get_parent_child_from_xpath(name):
    """Returns the parent and child elements from XPath."""
    if '/' in name:
        (pname, label) = name.rsplit('/', 1)
    else:
        pname = None
        label = name
    return (pname, label)


def to_bytes(in_str):
    """Converts a size to bytes.

    :param in_str: a number suffixed with a unit: {number}{unit}
                units supported:
                K, KB, k or kb - kilobytes
                M, MB, m or mb - megabytes
                G, GB, g or gb - gigabytes
                T, TB, t or tb - terabytes
    :returns: number of bytes
     None; if input is incorrect
    """
    match = re.search('^([0-9]+)([a-zA-Z]{0,2})$', in_str)

    if not match:
        return None

    unit = match.group(2).upper()
    value = match.group(1)

    size_count = int(value)
    if unit in ['K', 'KB']:
        multiplier = int(units.Ki)
    elif unit in ['M', 'MB']:
        multiplier = int(units.Mi)
    elif unit in ['G', 'GB']:
        multiplier = int(units.Gi)
    elif unit in ['T', 'TB']:
        multiplier = int(units.Ti)
    elif unit == "":
        return size_count
    else:
        return None

    size_in_bytes = int(size_count * multiplier)
    return size_in_bytes


def get_list(json_object, parent_node_name, child_node_name=None):
    """Returns a list of values from child_node_name.

    If child_node is not given, then it will retrieve list from parent node
    """
    if not json_object:
        return []

    return_list = []
    if isinstance(json_object[parent_node_name], list):
        for detail in json_object[parent_node_name]:
            if child_node_name:
                return_list.append(detail[child_node_name])
            else:
                return_list.append(detail)
    else:
        if child_node_name:
            return_list.append(json_object[parent_node_name][child_node_name])
        else:
            return_list.append(json_object[parent_node_name])

    return return_list


def get_node_value(json_object, parent_node_name, child_node_name=None):
    """Returns value of given child_node.

    If child_node is not given, then value of parent node is returned

    :returns: None If json_object or parent_node is not given,
                   If child_node is not found under parent_node
    """
    if not json_object:
        return None

    if not parent_node_name:
        return None

    detail = json_object[parent_node_name]
    if not child_node_name:
        return detail

    return_value = None

    if child_node_name in detail:
        return_value = detail[child_node_name]
    else:
        return_value = None

    return return_value


def format_err_msg_and_raise(operation_type, component,
                             error_message, error_code):
    """Method to format error message.

    :param operation_type: create, update, add, etc
    :param component: storagesystem, vpool, etc
    :param error_code: Error code from the API call
    :param error_message: Detailed error message
    """

    formated_err_msg = (_("Error: Failed to %(operation_type)s"
                          " %(component)s") %
                        {'operation_type': operation_type,
                         'component': component
                         })
    if error_message.startswith("\"\'") and error_message.endswith("\'\""):
        # stripping the first 2 and last 2 characters, which are quotes.
        error_message = error_message[2:len(error_message) - 2]

    formated_err_msg = formated_err_msg + "\nReason:" + error_message
    raise CoprHdError(error_code, formated_err_msg)


def search_by_tag(resource_search_uri, ipaddr, port):
    """Fetches the list of resources with a given tag.

    :param resource_search_uri: The tag based search uri
                              Example: '/block/volumes/search?tag=tagexample1'
    :param ipaddr: IP address of CoprHD host
    :param port: Port number
    """
    # check if the URI passed has both project and name parameters
    str_uri = six.text_type(resource_search_uri)
    if 'search' in str_uri and '?tag=' in str_uri:
        # Get the project URI

        (s, h) = service_json_request(
            ipaddr, port, "GET",
            resource_search_uri, None)

        o = json_decode(s)
        if not o:
            return None

        resources = get_node_value(o, "resource")

        resource_uris = []
        for resource in resources:
            resource_uris.append(resource["id"])
        return resource_uris
    else:
        raise CoprHdError(CoprHdError.VALUE_ERR, (_("Search URI %s"
                                                    " is not in the expected"
                                                    " format, it should end"
                                                    " with ?tag={0}")
                                                  % str_uri))


# Blocks the operation until the task is complete/error out/timeout
def block_until_complete(component_type,
                         resource_uri,
                         task_id,
                         ipaddr,
                         port,
                         synctimeout=0):

    if not synctimeout:
        synctimeout = TASK_TIMEOUT
    t = timeutils.StopWatch(duration=synctimeout)
    t.start()
    while not t.expired():
        if component_type == 'block':
            out = show_task_opid(task_id, ipaddr, port)
        else:
            out = get_task_by_resourceuri_and_taskId(
                component_type, resource_uri, task_id, ipaddr, port)

        if out:
            if out["state"] == "ready":

                # stop the timer and return
                t.stop()
                break

            # if the status of the task is 'error' then stop the timer
            # and raise exception
            if out["state"] == "error":
                # stop the timer
                t.stop()
                error_message = "Please see logs for more details"
                if ("service_error" in out and
                        "details" in out["service_error"]):
                    error_message = out["service_error"]["details"]
                raise CoprHdError(CoprHdError.VALUE_ERR,
                                  (_("Task: %(task_id)s"
                                     " is failed with"
                                     " error: %(error_message)s") %
                                   {'task_id': task_id,
                                    'error_message': error_message
                                    }))

    else:
        raise CoprHdError(CoprHdError.TIME_OUT,
                          (_("Task did not complete in %d secs."
                             " Operation timed out. Task in CoprHD"
                             " will continue") % synctimeout))

    return


def show_task_opid(taskid, ipaddr, port):
    (s, h) = service_json_request(
        ipaddr, port,
        "GET",
        URI_TASKS_BY_OPID.format(taskid),
        None)
    if (not s):
        return None
    o = json_decode(s)
    return o


def get_task_by_resourceuri_and_taskId(component_type, resource_uri,
                                       task_id, ipaddr, port):
    """Returns the single task details."""

    task_uri_constant = urihelper.singletonURIHelperInstance.getUri(
        component_type, "task")
    (s, h) = service_json_request(
        ipaddr, port, "GET",
        task_uri_constant.format(resource_uri, task_id), None)
    if not s:
        return None
    o = json_decode(s)
    return o


class CoprHdError(exception.VolumeBackendAPIException):

    """Custom exception class used to report logical errors.

    Attributes:
        err_code - String error code
        msg - String text
    """
    SOS_FAILURE_ERR = 1
    CMD_LINE_ERR = 2
    HTTP_ERR = 3
    VALUE_ERR = 4
    NOT_FOUND_ERR = 1
    ENTRY_ALREADY_EXISTS_ERR = 5
    MAX_COUNT_REACHED = 6
    TIME_OUT = 7

    def __init__(self, err_code, msg):
        self.err_code = err_code
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


class CoprHDResource(object):

    def __init__(self, ipaddr, port):
        """Constructor: takes IP address and port of the CoprHD instance.

        These are needed to make http requests for REST API
        """
        self.ipaddr = ipaddr
        self.port = port



# Copyright (c) 2015 Tintri.  All rights reserved.
# Copyright (c) 2012 NetApp, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Volume driver for Tintri storage.
"""

import datetime
import json
import math
import os
import re
import socket

from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import loopingcall
from oslo_utils import units
import requests
from six.moves import urllib

from cinder import exception
from cinder.i18n import _
from cinder.image import image_utils
from cinder import interface
from cinder import utils
from cinder.volume import configuration
from cinder.volume import driver
from cinder.volume.drivers import nfs

LOG = logging.getLogger(__name__)
default_api_version = 'v310'
img_prefix = 'image-'
tintri_path = '/tintri/'


tintri_opts = [
    cfg.StrOpt('tintri_server_hostname',
               help='The hostname (or IP address) for the storage system'),
    cfg.StrOpt('tintri_server_username',
               help='User name for the storage system'),
    cfg.StrOpt('tintri_server_password',
               help='Password for the storage system',
               secret=True),
    cfg.StrOpt('tintri_api_version',
               default=default_api_version,
               help='API version for the storage system'),
    cfg.IntOpt('tintri_image_cache_expiry_days',
               default=30,
               help='Delete unused image snapshots older than mentioned days'),
    cfg.StrOpt('tintri_image_shares_config',
               help='Path to image nfs shares file'),
]

CONF = cfg.CONF
CONF.register_opts(tintri_opts, group=configuration.SHARED_CONF_GROUP)


@interface.volumedriver
class TintriDriver(driver.ManageableVD,
                   driver.CloneableImageVD,
                   nfs.NfsDriver):
    """Base class for Tintri driver.

    Version History

    .. code-block:: none

        2.1.0.1 - Liberty driver
        2.2.0.1 - Mitaka driver
                -- Retype
                -- Image cache clean up
                -- Direct image clone fix
    """

    VENDOR = 'Tintri'
    VERSION = '2.2.0.1'
    # ThirdPartySystems wiki page
    CI_WIKI_NAME = "Tintri_CI"

    REQUIRED_OPTIONS = ['tintri_server_hostname', 'tintri_server_username',
                        'tintri_server_password']

    def __init__(self, *args, **kwargs):
        self._execute = None
        self._context = None
        super(TintriDriver, self).__init__(*args, **kwargs)
        self._execute_as_root = True
        self.configuration.append_config_values(tintri_opts)
        self.cache_cleanup = False
        self._mounted_image_shares = []

    def do_setup(self, context):
        self._image_shares_config = getattr(self.configuration,
                                            'tintri_image_shares_config')
        super(TintriDriver, self).do_setup(context)
        self._context = context
        self._check_ops(self.REQUIRED_OPTIONS, self.configuration)
        self._hostname = getattr(self.configuration, 'tintri_server_hostname')
        self._username = getattr(self.configuration, 'tintri_server_username')
        self._password = getattr(self.configuration, 'tintri_server_password')
        self._api_version = getattr(self.configuration, 'tintri_api_version')
        self._image_cache_expiry = getattr(self.configuration,
                                           'tintri_image_cache_expiry_days')

    def get_pool(self, volume):
        """Returns pool name where volume resides.

        :param volume: The volume hosted by the driver.
        :return: Name of the pool where given volume is hosted.
        """
        return volume['provider_location']

    def _get_client(self):
        """Returns a Tintri REST client connection."""
        return TClient(self._hostname, self._username, self._password,
                       self._api_version)

    def create_snapshot(self, snapshot):
        """Creates a snapshot."""
        (__, path) = self._get_export_ip_path(snapshot.volume_id)
        volume_path = '%s/%s' % (path, snapshot.volume_name)
        volume_path = '%(path)s/%(volume_name)s' % {
            'path': path,
            'volume_name': snapshot.volume_name,
        }
        model_update = {}
        with self._get_client() as c:
            provider_id = c.create_snapshot(volume_path,
                                            snapshot.volume.display_name or
                                            snapshot.volume_name,
                                            snapshot.volume_id,
                                            snapshot.display_name or
                                            snapshot.name)
            snapshot.provider_id = provider_id
            # Store Tintri snapshot ID as snapshot provider_id
            model_update['provider_id'] = provider_id

        return model_update

    def delete_snapshot(self, snapshot):
        """Deletes a snapshot."""
        if snapshot.provider_id:
            with self._get_client() as c:
                c.delete_snapshot(snapshot.provider_id)
        else:
            LOG.info('Snapshot %s not found', snapshot.name)

    def _check_ops(self, required_ops, configuration):
        """Ensures that the options we care about are set."""
        for op in required_ops:
            if not getattr(configuration, op):
                LOG.error('Configuration value %s is not set.', op)
                raise exception.InvalidConfigurationValue(option=op,
                                                          value=None)

    def create_volume_from_snapshot(self, volume, snapshot):
        """Creates a volume from snapshot."""
        vol_size = volume.size
        snap_size = snapshot.volume_size

        self._clone_snapshot(snapshot.provider_id, volume.name,
                             snapshot.volume_id)
        share = self._get_provider_location(snapshot.volume_id)
        volume['provider_location'] = share
        path = self.local_path(volume)

        self._set_rw_permissions(path)
        if vol_size != snap_size:
            try:
                self.extend_volume(volume, vol_size)
            except Exception:
                LOG.error('Resizing %s failed. Cleaning volume.',
                          volume.name)
                self._delete_file(path)
                raise

        return {'provider_location': volume['provider_location']}

    def _clone_snapshot(self, snapshot_id, clone_name, volume_id, share=None):
        """Clones volume from snapshot."""
        (host, path) = self._get_export_ip_path(volume_id, share)
        clone_path = '%s/%s-d' % (path, clone_name)
        with self._get_client() as c:
            c.clone_volume(snapshot_id, clone_path)

        self._move_cloned_volume(clone_name, volume_id, share)

    def _move_cloned_volume(self, clone_name, volume_id, share=None):
        local_path = self._get_local_path(volume_id, share)
        source_path = os.path.join(local_path, clone_name + '-d')
        if self._is_volume_present(source_path):
            source_file = os.listdir(source_path)[0]
            source = os.path.join(source_path, source_file)
            target = os.path.join(local_path, clone_name)
            moved = self._move_file(source, target)
            self._execute('rm', '-rf', source_path,
                          run_as_root=self._execute_as_root)
            if not moved:
                msg = (_('Failed to move volume %s.') % source)
                raise exception.VolumeDriverException(msg)
        else:
            raise exception.VolumeDriverException(
                _('Volume %s not found.') % source_path)

    def _clone_volume_to_volume(self, volume_name, clone_name,
                                volume_display_name, volume_id,
                                share=None, dst=None, image_id=None):
        """Creates volume snapshot then clones volume."""
        (__, path) = self._get_export_ip_path(volume_id, share)
        volume_path = '%s/%s' % (path, volume_name)
        if dst:
            (___, dst_path) = self._get_export_ip_path(None, dst)
            clone_path = '%s/%s-d' % (dst_path, clone_name)
        else:
            clone_path = '%s/%s-d' % (path, clone_name)
        with self._get_client() as c:
            if share and image_id:
                snapshot_id = self._create_image_snapshot(volume_name,
                                                          share,
                                                          image_id,
                                                          volume_display_name)
            else:
                snapshot_id = c.create_snapshot(
                    volume_path, volume_display_name, volume_id, volume_name,
                    deletion_policy='DELETE_ON_ZERO_CLONE_REFERENCES')
            c.clone_volume(snapshot_id, clone_path)

        self._move_cloned_volume(clone_name, volume_id, dst or share)

    @utils.synchronized('cache_cleanup')
    def _initiate_image_cache_cleanup(self):
        if self.cache_cleanup:
            LOG.debug('Image cache cleanup in progress.')
            return
        else:
            self.cache_cleanup = True
            timer = loopingcall.FixedIntervalLoopingCall(
                self._cleanup_cache)
            timer.start(interval=None)
            return timer

    def _cleanup_cache(self):
        LOG.debug('Cache cleanup: starting.')
        try:
            # Cleanup used cached image snapshots 30 days and older
            t = datetime.datetime.utcnow() - datetime.timedelta(
                days=self._image_cache_expiry)
            date = t.strftime("%Y-%m-%dT%H:%M:%S")
            with self._get_client() as c:
                # Get eligible snapshots to clean
                image_snaps = c.get_image_snapshots_to_date(date)
                if image_snaps:
                    for snap in image_snaps:
                        uuid = snap['uuid']['uuid']
                        LOG.debug(
                            'Cache cleanup: deleting image snapshot %s', uuid)
                        try:
                            c.delete_snapshot(uuid)
                        except Exception:
                            LOG.exception('Unexpected exception during '
                                          'cache cleanup of snapshot %s',
                                          uuid)
                else:
                    LOG.debug('Cache cleanup: nothing to clean')
        finally:
            self.cache_cleanup = False
            LOG.debug('Cache cleanup: finished')
            raise loopingcall.LoopingCallDone()

    def _update_volume_stats(self):
        """Retrieves stats info from volume group."""

        data = {}
        backend_name = self.configuration.safe_get('volume_backend_name')
        data['volume_backend_name'] = backend_name or self.VENDOR
        data['vendor_name'] = self.VENDOR
        data['driver_version'] = self.get_version()
        data['storage_protocol'] = self.driver_volume_type

        self._ensure_shares_mounted()
        self._initiate_image_cache_cleanup()
        pools = []
        for share in self._mounted_shares:
            pool = dict()
            capacity, free, used = self._get_capacity_info(share)
            pool['pool_name'] = share
            pool['total_capacity_gb'] = capacity / float(units.Gi)
            pool['free_capacity_gb'] = free / float(units.Gi)
            pool['reserved_percentage'] = 0
            pool['QoS_support'] = True
            pools.append(pool)
        data['pools'] = pools

        self._stats = data

    def _get_provider_location(self, volume_id):
        """Returns provider location for given volume."""
        volume = self.db.volume_get(self._context, volume_id)
        return volume.provider_location

    def _get_host_ip(self, volume_id):
        """Returns IP address for the given volume."""
        return self._get_provider_location(volume_id).split(':')[0]

    def _get_export_path(self, volume_id):
        """Returns NFS export path for the given volume."""
        return self._get_provider_location(volume_id).split(':')[1]

    def _resolve_hostname(self, hostname):
        """Resolves host name to IP address."""
        res = socket.getaddrinfo(hostname, None)[0]
        family, socktype, proto, canonname, sockaddr = res
        return sockaddr[0]

    def _is_volume_present(self, volume_path):
        """Checks if volume exists."""
        try:
            self._execute('ls', volume_path,
                          run_as_root=self._execute_as_root)
        except Exception:
            return False
        return True

    def _get_volume_path(self, nfs_share, volume_name):
        """Gets local volume path for given volume name on given nfs share."""
        return os.path.join(self._get_mount_point_for_share(nfs_share),
                            volume_name)

    def create_cloned_volume(self, volume, src_vref):
        """Creates a clone of the specified volume."""
        vol_size = volume.size
        src_vol_size = src_vref.size
        self._clone_volume_to_volume(src_vref.name, volume.name,
                                     src_vref.display_name,
                                     src_vref.id)

        share = self._get_provider_location(src_vref.id)
        volume['provider_location'] = share
        path = self.local_path(volume)

        self._set_rw_permissions(path)
        if vol_size != src_vol_size:
            try:
                self.extend_volume(volume, vol_size)
            except Exception:
                LOG.error('Resizing %s failed. Cleaning volume.',
                          volume.name)
                self._delete_file(path)
                raise

        return {'provider_location': volume['provider_location']}

    def copy_image_to_volume(self, context, volume, image_service, image_id):
        """Fetches the image from image_service and write it to the volume."""
        super(TintriDriver, self).copy_image_to_volume(
            context, volume, image_service, image_id)
        LOG.info('Copied image to volume %s using regular download.',
                 volume['name'])
        self._create_image_snapshot(volume['name'],
                                    volume['provider_location'], image_id,
                                    img_prefix + image_id)

    def _create_image_snapshot(self, volume_name, share, image_id, image_name):
        """Creates an image snapshot."""
        snapshot_name = img_prefix + image_id
        LOG.info('Creating image snapshot %s', snapshot_name)
        (host, path) = self._get_export_ip_path(None, share)
        volume_path = '%s/%s' % (path, volume_name)

        @utils.synchronized(snapshot_name, external=True)
        def _do_snapshot():
            with self._get_client() as c:
                snapshot_id = c.get_snapshot(image_id)
                if not snapshot_id:
                    snapshot_id = c.create_snapshot(volume_path, image_name,
                                                    image_id, snapshot_name)
                return snapshot_id

        try:
            return _do_snapshot()
        except Exception as e:
            LOG.warning('Exception while creating image %(image_id)s '
                        'snapshot. Exception: %(exc)s',
                        {'image_id': image_id, 'exc': e})

    def _find_image_snapshot(self, image_id):
        """Finds image snapshot."""
        with self._get_client() as c:
            return c.get_snapshot(image_id)

    def _clone_image_snapshot(self, snapshot_id, dst, share):
        """Clones volume from image snapshot."""
        file_path = self._get_volume_path(share, dst)
        if not os.path.exists(file_path):
            LOG.info('Cloning from snapshot to destination %s', dst)
            self._clone_snapshot(snapshot_id, dst, volume_id=None,
                                 share=share)

    def _delete_file(self, path):
        """Deletes file from disk and return result as boolean."""
        try:
            LOG.debug('Deleting file at path %s', path)
            cmd = ['rm', '-f', path]
            self._execute(*cmd, run_as_root=self._execute_as_root)
            return True
        except Exception as e:
            LOG.warning('Exception during deleting %s', e)
            return False

    def _move_file(self, source_path, dest_path):
        """Moves source to destination."""

        @utils.synchronized(dest_path, external=True)
        def _do_move(src, dst):
            if os.path.exists(dst):
                LOG.warning('Destination %s already exists.', dst)
                return False
            self._execute('mv', src, dst, run_as_root=self._execute_as_root)
            return True

        try:
            return _do_move(source_path, dest_path)
        except Exception as e:
            LOG.warning('Exception moving file %(src)s. Message: %(e)s',
                        {'src': source_path, 'e': e})
        return False

    def clone_image(self, context, volume,
                    image_location, image_meta,
                    image_service):
        """Creates a volume efficiently from an existing image.

        image_location is a string whose format depends on the
        image service backend in use. The driver should use it
        to determine whether cloning is possible.

        Returns a dict of volume properties eg. provider_location,
        boolean indicating whether cloning occurred.
        """
        image_name = image_meta['name']
        image_id = image_meta['id']
        if 'properties' in image_meta:
            provider_location = image_meta['properties'].get(
                'provider_location')
            if provider_location:
                image_location = (provider_location, None)
        cloned = False
        post_clone = False
        try:
            snapshot_id = self._find_image_snapshot(image_id)
            if snapshot_id:
                cloned = self._clone_from_snapshot(volume, image_id,
                                                   snapshot_id)
            else:
                cloned = self._direct_clone(volume, image_location,
                                            image_id, image_name)
            if cloned:
                post_clone = self._post_clone_image(volume)
        except Exception as e:
            LOG.info('Image cloning unsuccessful for image '
                     '%(image_id)s. Message: %(msg)s',
                     {'image_id': image_id, 'msg': e})
            vol_path = self.local_path(volume)
            volume['provider_location'] = None
            if os.path.exists(vol_path):
                self._delete_file(vol_path)
        finally:
            cloned = cloned and post_clone
            share = volume['provider_location'] if cloned else None
            bootable = True if cloned else False
            return {'provider_location': share, 'bootable': bootable}, cloned

    def _clone_from_snapshot(self, volume, image_id, snapshot_id):
        """Clones a copy from image snapshot."""
        cloned = False
        LOG.info('Cloning image %s from snapshot.', image_id)
        for share in self._mounted_shares:
            # Repeat tries in other shares if failed in some
            LOG.debug('Image share: %s', share)
            if (share and
                    self._is_share_vol_compatible(volume, share)):
                try:
                    self._clone_image_snapshot(snapshot_id, volume['name'],
                                               share)
                    cloned = True
                    volume['provider_location'] = share
                    break
                except Exception:
                    LOG.warning('Unexpected exception during '
                                'image cloning in share %s', share)
        return cloned

    def _direct_clone(self, volume, image_location, image_id, image_name):
        """Clones directly in nfs share."""
        LOG.info('Checking image clone %s from glance share.', image_id)
        cloned = False
        image_location = self._get_image_nfs_url(image_location)
        share = self._is_cloneable_share(image_location)
        run_as_root = self._execute_as_root

        dst_share = None
        for dst in self._mounted_shares:
            if dst and self._is_share_vol_compatible(volume, dst):
                dst_share = dst
                LOG.debug('Image dst share: %s', dst)
                break
        if not dst_share:
            return cloned

        LOG.debug('Share is cloneable %s', dst_share)
        volume['provider_location'] = dst_share
        (__, ___, img_file) = image_location.rpartition('/')
        dir_path = self._get_mount_point_for_share(share)
        dst_path = self._get_mount_point_for_share(dst_share)
        img_path = '%s/%s' % (dir_path, img_file)
        img_info = image_utils.qemu_img_info(img_path,
                                             run_as_root=run_as_root)
        if img_info.file_format == 'raw':
            LOG.debug('Image is raw %s', image_id)
            self._clone_volume_to_volume(
                img_file, volume['name'], image_name,
                volume_id=None, share=share, dst=dst_share, image_id=image_id)
            cloned = True
        else:
            LOG.info('Image will locally be converted to raw %s',
                     image_id)
            dst = '%s/%s' % (dst_path, volume['name'])
            image_utils.convert_image(img_path, dst, 'raw',
                                      run_as_root=run_as_root)
            data = image_utils.qemu_img_info(dst, run_as_root=run_as_root)
            if data.file_format != "raw":
                raise exception.InvalidResults(
                    _('Converted to raw, but '
                      'format is now %s') % data.file_format)
            else:
                cloned = True
                self._create_image_snapshot(
                    volume['name'], volume['provider_location'],
                    image_id, image_name)
        return cloned

    def _post_clone_image(self, volume):
        """Performs operations post image cloning."""
        LOG.info('Performing post clone for %s', volume['name'])
        vol_path = self.local_path(volume)
        self._set_rw_permissions(vol_path)
        self._resize_image_file(vol_path, volume['size'])
        return True

    def _resize_image_file(self, path, new_size):
        """Resizes the image file on share to new size."""
        LOG.debug('Checking file for resize.')
        if self._is_file_size_equal(path, new_size):
            return
        else:
            LOG.info('Resizing file to %sG', new_size)
            image_utils.resize_image(path, new_size,
                                     run_as_root=self._execute_as_root)
            if self._is_file_size_equal(path, new_size):
                return
            else:
                raise exception.InvalidResults(
                    _('Resizing image file failed.'))

    def _is_cloneable_share(self, image_location):
        """Finds if the image at location is cloneable."""
        conn, dr = self._check_nfs_path(image_location)
        return self._is_share_in_use(conn, dr)

    def _check_nfs_path(self, image_location):
        """Checks if the nfs path format is matched.

        WebNFS url format with relative-path is supported.
        Accepting all characters in path-names and checking against
        the mounted shares which will contain only allowed path segments.
        Returns connection and dir details.
        """
        conn, dr = None, None
        if image_location:
            nfs_loc_pattern = \
                r'^nfs://(([\w\-\.]+:[\d]+|[\w\-\.]+)(/[^/].*)*(/[^/\\\\]+))$'
            matched = re.match(nfs_loc_pattern, image_location)
            if not matched:
                LOG.debug('Image location not in the expected format %s',
                          image_location)
            else:
                conn = matched.group(2)
                dr = matched.group(3) or '/'
        return conn, dr

    def _is_share_in_use(self, conn, dr):
        """Checks if share is cinder mounted and returns it."""
        try:
            if conn:
                host = conn.split(':')[0]
                ip = self._resolve_hostname(host)
                for sh in self._mounted_shares + self._mounted_image_shares:
                    sh_ip = self._resolve_hostname(sh.split(':')[0])
                    sh_exp = sh.split(':')[1]
                    if sh_ip == ip and sh_exp == dr:
                        LOG.debug('Found share match %s', sh)
                        return sh
        except Exception:
            LOG.warning('Unexpected exception while listing used share.')

    def _get_image_nfs_url(self, image_location):
        """Gets direct url for nfs backend.

        It creates direct url from image_location
        which is a tuple with direct_url and locations.
        Returns url with nfs scheme if nfs store else returns url.
        It needs to be verified by backend before use.
        """

        direct_url, locations = image_location
        if not direct_url and not locations:
            raise exception.NotFound(_('Image location not present.'))

        # Locations will be always a list of one until
        # bp multiple-image-locations is introduced
        if not locations:
            return direct_url
        location = locations[0]
        url = location['url']
        if not location['metadata']:
            return url
        location_type = location['metadata'].get('type')
        if not location_type or location_type.lower() != "nfs":
            return url
        share_location = location['metadata'].get('share_location')
        mount_point = location['metadata'].get('mount_point')
        if not share_location or not mount_point:
            return url
        url_parse = urllib.parse.urlparse(url)
        abs_path = os.path.join(url_parse.netloc, url_parse.path)
        rel_path = os.path.relpath(abs_path, mount_point)
        direct_url = "%s/%s" % (share_location, rel_path)
        return direct_url

    def _is_share_vol_compatible(self, volume, share):
        """Checks if share is compatible with volume to host it."""
        return self._is_share_eligible(share, volume['size'])

    def _can_share_hold_size(self, share, size):
        """Checks if volume can hold image with size."""
        _tot_size, tot_available, _tot_allocated = self._get_capacity_info(
            share)
        if tot_available < size:
            msg = _('Container size smaller than required file size.')
            raise exception.VolumeDriverException(msg)

    def _get_export_ip_path(self, volume_id=None, share=None):
        """Returns export ip and path.

          One of volume id or share is used to return the values.
        """

        if volume_id:
            host_ip = self._get_host_ip(volume_id)
            export_path = self._get_export_path(volume_id)
        elif share:
            host_ip = share.split(':')[0]
            export_path = share.split(':')[1]
        else:
            raise exception.InvalidInput(
                reason=_('A volume ID or share was not specified.'))
        return host_ip, export_path

    def _get_local_path(self, volume_id=None, share=None):
        """Returns local path.

          One of volume id or share is used to return the values.
        """

        if volume_id:
            local_path = self._get_mount_point_for_share(
                self._get_provider_location(volume_id))
        elif share:
            local_path = self._get_mount_point_for_share(share)
        else:
            raise exception.InvalidInput(
                reason=_('A volume ID or share was not specified.'))
        return local_path

    def manage_existing(self, volume, existing_ref):
        """Brings an existing backend storage object under Cinder management.

        existing_ref is passed straight through from the API request's
        manage_existing_ref value, and it is up to the driver how this should
        be interpreted.  It should be sufficient to identify a storage object
        that the driver should somehow associate with the newly-created cinder
        volume structure.

        :param volume:       Cinder volume to manage
        :param existing_ref: Driver-specific information used to identify a
                             volume
        """
        nfs_share, nfs_mount, volume_name = self._get_share_mount(existing_ref)

        LOG.debug('Managing volume %(vol)s with ref %(ref)s',
                  {'vol': volume['id'], 'ref': existing_ref})
        if volume_name != volume['name']:
            src = os.path.join(nfs_mount, volume_name)
            dst = os.path.join(nfs_mount, volume['name'])
            if not self._move_file(src, dst):
                msg = (_('Failed to manage volume %s.') %
                       existing_ref['source-name'])
                raise exception.VolumeDriverException(msg)
            self._set_rw_permissions(dst)

        LOG.info('Manage volume %s', volume['name'])
        return {'provider_location': nfs_share}

    def manage_existing_get_size(self, volume, existing_ref):
        """Returns size of volume to be managed by manage_existing.

        When calculating the size, round up to the next GB.

        :param volume:       Cinder volume to manage
        :param existing_ref: Driver-specific information used to identify a
                             volume
        """
        nfs_share, nfs_mount, volume_name = self._get_share_mount(existing_ref)

        try:
            volume_path = os.path.join(nfs_mount, volume_name)
            vol_size = int(math.ceil(float(utils.get_file_size(volume_path)) /
                                     units.Gi))
        except OSError:
            msg = (_('Failed to get size of volume %s') %
                   existing_ref['source-name'])
            raise exception.VolumeDriverException(msg)

        return vol_size

    def unmanage(self, volume):
        """Removes the specified volume from Cinder management.

        Does not delete the underlying backend storage object.

        :param volume: Cinder volume to unmanage
        """
        volume_path = self.local_path(volume)
        LOG.info('Unmanage volume %s', volume_path)

    def _convert_volume_share(self, volume_share):
        """Converts the share name to IP address."""
        share_split = volume_share.rsplit(':', 1)
        return self._resolve_hostname(share_split[0]) + ':' + share_split[1]

    def _get_share_mount(self, vol_ref):
        """Get the NFS share, NFS mount, and volume path from reference.

        :param vol_ref: Driver-specific information used to identify a volume
        :return:        NFS Share, NFS mount, volume path
        """
        if 'source-name' not in vol_ref or not vol_ref['source-name']:
            msg = _('Volume reference must contain source-name element.')
            raise exception.ManageExistingInvalidReference(
                existing_ref=vol_ref, reason=msg)

        volume_share = self._convert_volume_share(vol_ref['source-name'])
        for nfs_share in self._mounted_shares:
            share = self._convert_volume_share(nfs_share)
            (__, match, volume_name) = volume_share.partition(share)
            if match == share:
                volume_name = volume_name.lstrip('/')
                nfs_mount = self._get_mount_point_for_share(nfs_share)
                volume_path = os.path.join(nfs_mount, volume_name)
                if os.path.isfile(volume_path):
                    LOG.debug('Found volume %(path)s on share %(share)s',
                              {'path': volume_path, 'share': nfs_share})
                    return nfs_share, nfs_mount, volume_name
                else:
                    LOG.debug('Volume ref %(ref)s not on share %(share)s',
                              {'ref': vol_ref, 'share': nfs_share})

        raise exception.ManageExistingInvalidReference(
            existing_ref=vol_ref, reason=_('Volume not found.'))

    def retype(self, context, volume, new_type, diff, host):
        """Retype from one volume type to another.

        At this point Tintri VMstore does not differentiate between
        volume types on the same array. This is a no-op for us.
        """
        return True, None

    def _ensure_shares_mounted(self):
        # Mount image shares, we do not need to store these mounts
        # in _mounted_shares
        mounted_image_shares = []
        if self._image_shares_config:
            self._load_shares_config(self._image_shares_config)
            for share in self.shares.keys():
                try:
                    self._ensure_share_mounted(share)
                    mounted_image_shares.append(share)
                except Exception:
                    LOG.exception('Exception during mounting.')
        self._mounted_image_shares = mounted_image_shares

        # Mount Cinder shares
        super(TintriDriver, self)._ensure_shares_mounted()


class TClient(object):
    """REST client for Tintri storage."""

    def __init__(self, hostname, username, password,
                 api_version=default_api_version):
        """Initializes a connection to Tintri server."""
        self.api_url = 'https://' + hostname + '/api'
        self.api_version = api_version
        self.session_id = self.login(username, password)
        self.headers = {'content-type': 'application/json',
                        'cookie': 'JSESSIONID=' + self.session_id}

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.logout()

    def get(self, api):
        return self.get_query(api, None)

    def get_query(self, api, query):
        url = self.api_url + api

        return requests.get(url, headers=self.headers,
                            params=query, verify=False)

    def delete(self, api):
        url = self.api_url + api

        return requests.delete(url, headers=self.headers, verify=False)

    def put(self, api, payload):
        url = self.api_url + api

        return requests.put(url, data=json.dumps(payload),
                            headers=self.headers, verify=False)

    def post(self, api, payload):
        url = self.api_url + api

        return requests.post(url, data=json.dumps(payload),
                             headers=self.headers, verify=False)

    def login(self, username, password):
        # Payload, header and URL for login
        headers = {'content-type': 'application/json',
                   'Tintri-Api-Client':
                   'Tintri-Cinder-Driver-%s' % TintriDriver.VERSION}
        payload = {'username': username,
                   'password': password,
                   'typeId': 'com.tintri.api.rest.vcommon.dto.rbac.'
                             'RestApiCredentials'}
        url = self.api_url + '/' + self.api_version + '/session/login'

        r = requests.post(url, data=json.dumps(payload),
                          headers=headers, verify=False)

        if r.status_code != 200:
            msg = _('Failed to login for user %s.') % username
            raise exception.VolumeDriverException(msg)

        return r.cookies['JSESSIONID']

    def logout(self):
        url = self.api_url + '/' + self.api_version + '/session/logout'

        requests.get(url, headers=self.headers, verify=False)

    @staticmethod
    def _remove_prefix(volume_path, prefix):
        if volume_path.startswith(prefix):
            return volume_path[len(prefix):]
        else:
            return volume_path

    def create_snapshot(self, volume_path, volume_name, volume_id,
                        snapshot_name, deletion_policy=None):
        """Creates a volume snapshot."""
        request = {'typeId': 'com.tintri.api.rest.' + self.api_version +
                             '.dto.domain.beans.cinder.CinderSnapshotSpec',
                   'file': TClient._remove_prefix(volume_path, tintri_path),
                   'vmName': volume_name or snapshot_name,
                   'description': snapshot_name + ' (' + volume_id + ')',
                   'vmTintriUuid': volume_id,
                   'instanceId': volume_id,
                   'snapshotCreator': 'Cinder',
                   'deletionPolicy': deletion_policy,
                   }

        payload = '/' + self.api_version + '/cinder/snapshot'
        r = self.post(payload, request)
        if r.status_code != 200:
            msg = _('Failed to create snapshot for volume %s.') % volume_path
            raise exception.VolumeDriverException(msg)

        return r.json()[0]

    def get_snapshot(self, volume_id):
        """Gets a volume snapshot."""
        filter = {'vmUuid': volume_id}

        payload = '/' + self.api_version + '/snapshot'
        r = self.get_query(payload, filter)
        if r.status_code != 200:
            msg = _('Failed to get snapshot for volume %s.') % volume_id
            raise exception.VolumeDriverException(msg)

        if int(r.json()['filteredTotal']) > 0:
            return r.json()['items'][0]['uuid']['uuid']

    def get_image_snapshots_to_date(self, date):
        filter = {'sortedBy': 'createTime',
                  'target': 'SNAPSHOT',
                  'consistency': 'CRASH_CONSISTENT',
                  'hasClone': 'No',
                  'type': 'CINDER_GENERATED_SNAPSHOT',
                  'contain': 'image-',
                  'limit': '100',
                  'page': '1',
                  'sortOrder': 'DESC',
                  'since': '1970-01-01T00:00:00',
                  'until': date,
                  }
        payload = '/' + self.api_version + '/snapshot'
        r = self.get_query(payload, filter)
        if r.status_code != 200:
            msg = _('Failed to get image snapshots.')
            raise exception.VolumeDriverException(msg)
        return r.json()['items']

    def delete_snapshot(self, snapshot_uuid):
        """Deletes a snapshot."""
        url = '/' + self.api_version + '/snapshot/'
        self.delete(url + snapshot_uuid)

    def clone_volume(self, snapshot_uuid, volume_path):
        """Clones a volume from snapshot."""
        request = {'typeId': 'com.tintri.api.rest.' + self.api_version +
                             '.dto.domain.beans.cinder.CinderCloneSpec',
                   'destinationPaths':
                       [TClient._remove_prefix(volume_path, tintri_path)],
                   'tintriSnapshotUuid': snapshot_uuid,
                   }

        url = '/' + self.api_version + '/cinder/clone'
        r = self.post(url, request)
        if r.status_code != 200 and r.status_code != 204:
            msg = _('Failed to clone volume from snapshot %s.') % snapshot_uuid
            raise exception.VolumeDriverException(msg)



# Copyright (c) 2016 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import hashlib
import json
import six
import time

from oslo_log import log as logging
from oslo_service import loopingcall
from oslo_utils import units

from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder.volume.drivers.huawei import constants
from cinder.volume import utils

LOG = logging.getLogger(__name__)


def encode_name(id):
    encoded_name = hashlib.md5(id.encode('utf-8')).hexdigest()
    prefix = id.split('-')[0] + '-'
    postfix = encoded_name[:constants.MAX_NAME_LENGTH - len(prefix)]
    return prefix + postfix


def old_encode_name(id):
    pre_name = id.split("-")[0]
    vol_encoded = six.text_type(hash(id))
    if vol_encoded.startswith('-'):
        newuuid = pre_name + vol_encoded
    else:
        newuuid = pre_name + '-' + vol_encoded
    return newuuid


def encode_host_name(name):
    if name and len(name) > constants.MAX_NAME_LENGTH:
        encoded_name = hashlib.md5(name.encode('utf-8')).hexdigest()
        return encoded_name[:constants.MAX_NAME_LENGTH]
    return name


def old_encode_host_name(name):
    if name and len(name) > constants.MAX_NAME_LENGTH:
        name = six.text_type(hash(name))
    return name


def wait_for_condition(func, interval, timeout):
    start_time = time.time()

    def _inner():
        try:
            res = func()
        except Exception as ex:
            raise exception.VolumeBackendAPIException(data=ex)

        if res:
            raise loopingcall.LoopingCallDone()

        if int(time.time()) - start_time > timeout:
            msg = (_('wait_for_condition: %s timed out.')
                   % func.__name__)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    timer = loopingcall.FixedIntervalLoopingCall(_inner)
    timer.start(interval=interval).wait()


def get_volume_size(volume):
    """Calculate the volume size.

    We should divide the given volume size by 512 for the 18000 system
    calculates volume size with sectors, which is 512 bytes.
    """
    volume_size = units.Gi / 512  # 1G
    if int(volume.size) != 0:
        volume_size = int(volume.size) * units.Gi / 512

    return volume_size


def get_volume_metadata(volume):
    if type(volume) is objects.Volume:
        return volume.metadata

    if 'volume_metadata' in volume:
        metadata = volume.get('volume_metadata')
        return {item['key']: item['value'] for item in metadata}

    return {}


def get_admin_metadata(volume):
    admin_metadata = {}
    if 'admin_metadata' in volume:
        admin_metadata = volume.admin_metadata
    elif 'volume_admin_metadata' in volume:
        metadata = volume.get('volume_admin_metadata', [])
        admin_metadata = {item['key']: item['value'] for item in metadata}

    LOG.debug("Volume ID: %(id)s, admin_metadata: %(admin_metadata)s.",
              {"id": volume.id, "admin_metadata": admin_metadata})
    return admin_metadata


def get_snapshot_metadata_value(snapshot):
    if type(snapshot) is objects.Snapshot:
        return snapshot.metadata

    if 'snapshot_metadata' in snapshot:
        metadata = snapshot.snapshot_metadata
        return {item['key']: item['value'] for item in metadata}

    return {}


def check_whether_operate_consistency_group(func):
    def wrapper(self, context, group, *args, **kwargs):
        if not utils.is_group_a_cg_snapshot_type(group):
            msg = _("%s, the group or group snapshot is not cg or "
                    "cg_snapshot") % func.__name__
            LOG.debug(msg)
            raise NotImplementedError(msg)
        return func(self, context, group, *args, **kwargs)
    return wrapper


def to_string(**kwargs):
    return json.dumps(kwargs) if kwargs else ''


def get_lun_metadata(volume):
    if not volume.provider_location:
        return {}

    info = json.loads(volume.provider_location)
    if isinstance(info, dict):
        return info

    # To keep compatible with old driver version
    admin_metadata = get_admin_metadata(volume)
    metadata = get_volume_metadata(volume)
    return {'huawei_lun_id': six.text_type(info),
            'huawei_lun_wwn': admin_metadata.get('huawei_lun_wwn'),
            'hypermetro_id': metadata.get('hypermetro_id'),
            'remote_lun_id': metadata.get('remote_lun_id')
            }


def get_snapshot_metadata(snapshot):
    if not snapshot.provider_location:
        return {}

    info = json.loads(snapshot.provider_location)
    if isinstance(info, dict):
        return info

    # To keep compatible with old driver version
    return {'huawei_snapshot_id': six.text_type(info)}


def get_volume_lun_id(client, volume):
    metadata = get_lun_metadata(volume)
    lun_id = metadata.get('huawei_lun_id')

    # First try the new encoded way.
    if not lun_id:
        volume_name = encode_name(volume.id)
        lun_id = client.get_lun_id_by_name(volume_name)

    # If new encoded way not found, try the old encoded way.
    if not lun_id:
        volume_name = old_encode_name(volume.id)
        lun_id = client.get_lun_id_by_name(volume_name)

    return lun_id, metadata.get('huawei_lun_wwn')


def get_snapshot_id(client, snapshot):
    metadata = get_snapshot_metadata(snapshot)
    snapshot_id = metadata.get('huawei_snapshot_id')

    # First try the new encoded way.
    if not snapshot_id:
        name = encode_name(snapshot.id)
        snapshot_id = client.get_snapshot_id_by_name(name)

    # If new encoded way not found, try the old encoded way.
    if not snapshot_id:
        name = old_encode_name(snapshot.id)
        snapshot_id = client.get_snapshot_id_by_name(name)

    return snapshot_id


def get_host_id(client, host_name):
    encoded_name = encode_host_name(host_name)
    host_id = client.get_host_id_by_name(encoded_name)
    if encoded_name == host_name:
        return host_id

    if not host_id:
        encoded_name = old_encode_host_name(host_name)
        host_id = client.get_host_id_by_name(encoded_name)

    return host_id



# Copyright (c) 2014 Navneet Singh.  All rights reserved.
# Copyright (c) 2014 Clinton Knight.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Utilities for NetApp E-series drivers.
"""

import base64
import binascii
import uuid

import six


MULTI_ATTACH_HOST_GROUP_NAME = 'cinder-multi-attach'
NULL_REF = '0000000000000000000000000000000000000000'
MAX_LUNS_PER_HOST = 256
MAX_LUNS_PER_HOST_GROUP = 256


def encode_hex_to_base32(hex_string):
    """Encodes hex to base32 bit as per RFC4648."""
    bin_form = binascii.unhexlify(hex_string)
    return base64.b32encode(bin_form)


def decode_base32_to_hex(base32_string):
    """Decodes base32 string to hex string."""
    bin_form = base64.b32decode(base32_string)
    return binascii.hexlify(bin_form)


def convert_uuid_to_es_fmt(uuid_str):
    """Converts uuid to e-series compatible name format."""
    uuid_base32 = encode_hex_to_base32(uuid.UUID(six.text_type(uuid_str)).hex)
    es_label = uuid_base32.strip(b'=')
    if six.PY3:
        es_label = es_label.decode('ascii')
    return es_label


def convert_es_fmt_to_uuid(es_label):
    """Converts e-series name format to uuid."""
    if isinstance(es_label, six.text_type):
        es_label = es_label.encode('utf-8')
    if es_label.startswith(b'tmp-'):
        es_label = es_label[4:]
    es_label = es_label.ljust(32, b'=')
    es_label = binascii.hexlify(base64.b32decode(es_label))
    if six.PY3:
        es_label = es_label.decode('ascii')
    return uuid.UUID(es_label)



# Copyright (c) 2016 EMC Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import oslo_serialization

from cinder.i18n import _
from cinder.volume.drivers.coprhd.helpers import commoncoprhdapi as common
from cinder.volume.drivers.coprhd.helpers import consistencygroup
from cinder.volume.drivers.coprhd.helpers import volume


class Snapshot(common.CoprHDResource):

    # Commonly used URIs for the 'Snapshot' module
    URI_SNAPSHOTS = '/{0}/snapshots/{1}'
    URI_BLOCK_SNAPSHOTS = '/block/snapshots/{0}'
    URI_SEARCH_SNAPSHOT_BY_TAG = '/block/snapshots/search?tag={0}'
    URI_SNAPSHOT_LIST = '/{0}/{1}/{2}/protection/snapshots'
    URI_SNAPSHOT_TASKS_BY_OPID = '/vdc/tasks/{0}'
    URI_RESOURCE_DEACTIVATE = '{0}/deactivate'
    URI_CONSISTENCY_GROUP = "/block/consistency-groups"
    URI_CONSISTENCY_GROUPS_SNAPSHOT_INSTANCE = (
        URI_CONSISTENCY_GROUP + "/{0}/protection/snapshots/{1}")
    URI_CONSISTENCY_GROUPS_SNAPSHOT_DEACTIVATE = (
        URI_CONSISTENCY_GROUPS_SNAPSHOT_INSTANCE + "/deactivate")
    URI_BLOCK_SNAPSHOTS_TAG = URI_BLOCK_SNAPSHOTS + '/tags'

    VOLUMES = 'volumes'
    CG = 'consistency-groups'
    BLOCK = 'block'

    timeout = 300

    def snapshot_list_uri(self, otype, otypename, ouri):
        """Makes REST API call to list snapshots under a volume.

        :param otype: block
        :param otypename: either volume or consistency-group should be
                          provided
        :param ouri: uri of volume or consistency-group
        :returns: list of snapshots
        """
        (s, h) = common.service_json_request(
            self.ipaddr, self.port,
            "GET",
            Snapshot.URI_SNAPSHOT_LIST.format(otype, otypename, ouri), None)
        o = common.json_decode(s)
        return o['snapshot']

    def snapshot_show_uri(self, otype, resource_uri, suri):
        """Retrieves snapshot details based on snapshot Name or Label.

        :param otype: block
        :param suri: uri of the Snapshot.
        :param resource_uri: uri of the source resource
        :returns: Snapshot details in JSON response payload
        """
        if(resource_uri is not None and
           resource_uri.find('BlockConsistencyGroup') > 0):
            (s, h) = common.service_json_request(
                self.ipaddr, self.port,
                "GET",
                Snapshot.URI_CONSISTENCY_GROUPS_SNAPSHOT_INSTANCE.format(
                    resource_uri,
                    suri),
                None)
        else:
            (s, h) = common.service_json_request(
                self.ipaddr, self.port,
                "GET",
                Snapshot.URI_SNAPSHOTS.format(otype, suri), None)

        return common.json_decode(s)

    def snapshot_query(self, storageres_type,
                       storageres_typename, resuri, snapshot_name):
        if resuri is not None:
            uris = self.snapshot_list_uri(
                storageres_type,
                storageres_typename,
                resuri)
            for uri in uris:
                snapshot = self.snapshot_show_uri(
                    storageres_type,
                    resuri,
                    uri['id'])
                if (False == common.get_node_value(snapshot, 'inactive') and
                        snapshot['name'] == snapshot_name):
                    return snapshot['id']

        raise common.CoprHdError(
            common.CoprHdError.SOS_FAILURE_ERR,
            (_("snapshot with the name: "
               "%s Not Found") % snapshot_name))

    def storage_resource_query(self,
                               storageres_type,
                               volume_name,
                               cg_name,
                               project,
                               tenant):
        resourcepath = "/" + project
        if tenant is not None:
            resourcepath = tenant + resourcepath

        resUri = None
        resourceObj = None
        if Snapshot.BLOCK == storageres_type and volume_name is not None:
            resourceObj = volume.Volume(self.ipaddr, self.port)
            resUri = resourceObj.volume_query(resourcepath, volume_name)
        elif Snapshot.BLOCK == storageres_type and cg_name is not None:
            resourceObj = consistencygroup.ConsistencyGroup(
                self.ipaddr,
                self.port)
            resUri = resourceObj.consistencygroup_query(
                cg_name,
                project,
                tenant)
        else:
            resourceObj = None

        return resUri

    def snapshot_create(self, otype, typename, ouri,
                        snaplabel, inactive, sync,
                        readonly=False, synctimeout=0):
        """New snapshot is created, for a given volume.

        :param otype: block type should be provided
        :param typename: either volume or consistency-groups should
                         be provided
        :param ouri: uri of volume
        :param snaplabel: name of the snapshot
        :param inactive: if true, the snapshot will not activate the
                         synchronization between source and target volumes
        :param sync: synchronous request
        :param synctimeout: Query for task status for 'synctimeout' secs.
                            If the task doesn't complete in synctimeout secs,
                            an exception is thrown
        """

        # check snapshot is already exist
        is_snapshot_exist = True
        try:
            self.snapshot_query(otype, typename, ouri, snaplabel)
        except common.CoprHdError as e:
            if e.err_code == common.CoprHdError.NOT_FOUND_ERR:
                is_snapshot_exist = False
            else:
                raise

        if is_snapshot_exist:
            raise common.CoprHdError(
                common.CoprHdError.ENTRY_ALREADY_EXISTS_ERR,
                (_("Snapshot with name %(snaplabel)s"
                   " already exists under %(typename)s") %
                 {'snaplabel': snaplabel,
                  'typename': typename
                  }))

        parms = {
            'name': snaplabel,
            # if true, the snapshot will not activate the synchronization
            # between source and target volumes
            'create_inactive': inactive
        }
        if readonly is True:
            parms['read_only'] = readonly
        body = oslo_serialization.jsonutils.dumps(parms)

        # REST api call
        (s, h) = common.service_json_request(
            self.ipaddr, self.port,
            "POST",
            Snapshot.URI_SNAPSHOT_LIST.format(otype, typename, ouri), body)
        o = common.json_decode(s)

        task = o["task"][0]

        if sync:
            return (
                common.block_until_complete(
                    otype,
                    task['resource']['id'],
                    task["id"], self.ipaddr, self.port, synctimeout)
            )
        else:
            return o

    def snapshot_delete_uri(self, otype, resource_uri,
                            suri, sync, synctimeout=0):
        """Delete a snapshot by uri.

        :param otype: block
        :param resource_uri: uri of the source resource
        :param suri: Uri of the Snapshot
        :param sync: To perform operation synchronously
        :param synctimeout: Query for task status for 'synctimeout' secs. If
                            the task doesn't complete in synctimeout secs,
                            an exception is thrown
        """
        s = None
        if resource_uri.find("Volume") > 0:

            (s, h) = common.service_json_request(
                self.ipaddr, self.port,
                "POST",
                Snapshot.URI_RESOURCE_DEACTIVATE.format(
                    Snapshot.URI_BLOCK_SNAPSHOTS.format(suri)),
                None)
        elif resource_uri.find("BlockConsistencyGroup") > 0:

            (s, h) = common.service_json_request(
                self.ipaddr, self.port,
                "POST",
                Snapshot.URI_CONSISTENCY_GROUPS_SNAPSHOT_DEACTIVATE.format(
                    resource_uri,
                    suri),
                None)
        o = common.json_decode(s)
        task = o["task"][0]

        if sync:
            return (
                common.block_until_complete(
                    otype,
                    task['resource']['id'],
                    task["id"], self.ipaddr, self.port, synctimeout)
            )
        else:
            return o

    def snapshot_delete(self, storageres_type,
                        storageres_typename, resource_uri,
                        name, sync, synctimeout=0):
        snapshotUri = self.snapshot_query(
            storageres_type,
            storageres_typename,
            resource_uri,
            name)
        self.snapshot_delete_uri(
            storageres_type,
            resource_uri,
            snapshotUri,
            sync, synctimeout)
